/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define.async('io.ox/office/tk/config',
    ['io.ox/core/gettext',
     'settings!io.ox/office'
    ], function (GetText, settings) {

    'use strict';

    var // the configuration items of the OX Documents modules
        documentsConfig = settings.get('module', {});

    // static class Config ====================================================

    /**
     * Wraps the configuration of the OX Documents module.
     */
    var Config = {};

    // methods ----------------------------------------------------------------

    /**
     * Returns the value of a specific configuration property, or the complete
     * set of all available configuration properties.
     *
     * @param {String} [key]
     *  If specified, the key of a specific configuration property. If omitted,
     *  this method will return the entire configuration as a map.
     *
     * @param {Any} [def]
     *  A default value, if the configuration property specified with the 'key'
     *  parameter is not set.
     *
     * @returns {Any}
     *  The value of a specific configuration property, or the complete set of
     *  configuration properties of the OX Documents module, according to the
     *  'key' parameter.
     */
    Config.get = function (key, def) {
        return _.isString(key) ? ((key in documentsConfig) ? documentsConfig[key] : def) : documentsConfig;
    };

    /**
     * Returns the state of a specific Boolean configuration property.
     *
     * @param {String} key
     *  The key of a specific Boolean configuration property.
     *
     * @param {Boolean} [def=false]
     *  The default state, if the configuration property specified with the
     *  'key' parameter is not set.
     *
     * @returns {Boolean}
     *  The state of the specified configuration property. If the property does
     *  not exist, this method will return the default value.
     */
    Config.getFlag = function (key, def) {
        return Config.get(key, def === true) === true;
    };

    /**
     * Returns the state of the specified Boolean flag in the anchor part of
     * the browser URL.
     *
     * @param {String} flag
     *  The name of the flag.
     *
     * @param {Boolean} [def=false]
     *  The default state, if the URL does not contain the specified option.
     *
     * @returns {Boolean}
     *  The state of the specified flag in the anchor part of the URL.
     */
    Config.getUrlFlag = function (flag, def) {
        var value = _.url.hash(flag);
        return (/^(1|true)$/i).test(value) || ((/^(0|false)$/i).test(value) ? false : (def === true));
    };

    /**
     * Returns whether a specific debug mode is enabled in all OX Documents
     * applications. the debug mode will be enabled, if the passed server-side
     * configuration property is set to true, and the passed URL hash option is
     * NOT explicitly set to false; or if the server-side configuration
     * property 'debugondemand' is set to true, and the passed URL hash option
     * is set to true.
     *
     * @param {String} configFlag
     *  The name of the server-side configuration item.
     *
     * @param {String} urlFlag
     *  The name of the URL hash option.
     *
     * @returns {Boolean}
     *  Whether the specified debug mode is enabled.
     */
    Config.getDebugState = function (configFlag, urlFlag) {
        return Config.getFlag(configFlag) ? Config.getUrlFlag(urlFlag, true) : (Config.getFlag('debugondemand') && Config.getUrlFlag(urlFlag));
    };

    // constants --------------------------------------------------------------

    /**
     * The full identifier of the current locale, with leading lower-case
     * language identifier, and trailing upper-case country identifier,
     * separated by an underscore character, e.g. 'en_US'.
     *
     * @constant
     */
    Config.LOCALE = '';

    /**
     * The lower-case language identifier of the current locale, e.g. 'en'.
     *
     * @constant
     */
    Config.LANGUAGE = '';

    /**
     * The upper-case country identifier of the current locale, e.g. 'US'.
     *
     * @constant
     */
    Config.COUNTRY = '';

    /**
     * A Boolean flag specifying whether the debug mode is enabled in all OX
     * Documents applications. Debug mode will be enabled, if the server-side
     * configuration property 'debugavailable' is set to true, and the URL hash
     * option 'office:enable-debug' is NOT explicitly set to false; or if the
     * server-side configuration property 'debugondemand' is set to true, and
     * the URL hash option 'office:enable-debug' is set to true.
     *
     * @constant
     */
    Config.DEBUG = Config.getDebugState('debugavailable', 'office:enable-debug');

    // user data ==============================================================

    // get locale data of the user currently logged in
    var userPromise = GetText.language.done(function (locale) {

        var // language and country identifier from the locale
            matches = /^([a-z]+)(_([A-Z]+))?/.exec(locale);

        // locale settings
        if (_.isArray(matches)) {
            Config.LOCALE = locale;
            Config.LANGUAGE = matches[1] || '';
            Config.COUNTRY = matches[3] || '';
        } else {
            window.console.error('Config: unrecognized locale "' + locale + '"');
        }
    });

    // exports ================================================================

    // create the exported promise that will be resolved with the Config class
    return userPromise.then(_.constant(Config), _.constant(Config));

});
