/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/tk/control/textfield',
    ['io.ox/office/tk/utils',
     'io.ox/office/tk/keycodes',
     'io.ox/office/tk/forms',
     'io.ox/office/tk/object/baseobject',
     'io.ox/office/tk/control/group',
     'io.ox/office/tk/control/widthmixin'
    ], function (Utils, KeyCodes, Forms, BaseObject, Group, WidthMixin) {

    'use strict';

    var // default validator without any restrictions on the field text
        defaultValidator = null;

    // class TextField ========================================================

    /**
     * Creates a container element used to hold a text input field.
     *
     * @constructor
     *
     * @extends Group
     * @extends WidthMixin
     *
     * @param {Object} [initOptions]
     *  Optional parameters. Supports all options of the Group base class
     *  constructor, all formatting options for input elements supported by the
     *  method Forms.createInputMarkup(), and all options of the mix-in class
     *  WidthMixin. Additionally, the following options are supported:
     *  @param {Boolean} [initOptions.select=false]
     *      If set to true, the entire text will be selected after the text
     *      field has been clicked, or if the method TextField.grabFocus() has
     *      been called. The text will always be selected (independently from
     *      this option), if the text field gets focus via keyboard navigation.
     *  @param {TextField.Validator} [initOptions.validator]
     *      A text validator that will be used to convert the values from the
     *      control values (passed to the Group.setValue() method) to the text
     *      representation used in this text field, to validate the text while
     *      typing in the text field, and to convert the entered text to the
     *      control value (returned by the Group.getValue() method). If no
     *      validator has been specified, a default validator will be used that
     *      does not perform any conversions.
     *  @param {Object} [initOptions.smallerVersion]
     *      If specified, the Textfield will known, how it should act, if there
     *      will be not enough free place to show normal view
     */
    function TextField(initOptions) {

        var // self reference
            self = this,

            // the <input> element
            inputNode = $(Forms.createInputMarkup(initOptions)),

            // create the wrapper node for the input field (IE renders input fields with padding wrong)
            wrapperNode = $('<div>').addClass('input-wrapper').append(inputNode),

            // whether to select the entire text on click
            select = Utils.getBooleanOption(initOptions, 'select', false),

            // the validator used to convert and validate values
            validator = Utils.getObjectOption(initOptions, 'validator', defaultValidator),

            // saved state of the text field, used to restore while validating
            inputState = null,

            // initial value of text field when focused, needed for ESCAPE key handling
            initialText = null,

            // whether the group is focused (needed to detect whether to restore selection after focus change inside the group)
            groupFocused = false,

            // saves the (old) styles in case of switching to small version
            savedStyles = {},

            // smaller version of group (to fit in smaller resolutions)
            smallerVersion = Utils.getOption(initOptions, 'smallerVersion', false);

        // base constructor ---------------------------------------------------

        Group.call(this, initOptions);
        WidthMixin.call(this, wrapperNode, initOptions);

        // private methods ----------------------------------------------------

        /**
         * Saves the current value and selection of the text field node.
         */
        function saveInputState() {
            inputState = Forms.getInputSelection(inputNode);
            inputState.text = inputNode.val();
        }

        /**
         * Restores the current value and selection of the text field.
         */
        function restoreInputState() {
            inputNode.val(inputState.text);
            Forms.setInputSelection(inputNode, inputState.start, inputState.end);
        }

        /**
         * The update handler for this text field.
         */
        function updateHandler(value) {
            value = _.isNull(value) ? '' : validator.valueToText(value);
            // do not access the input element if the value does not change (browsers will destroy current selection)
            if (inputNode.val() !== value) {
                inputNode.val(value);
                self.getNode().attr('data-value', value);
                saveInputState();
            }
        }

        /**
         * Handles all focus events of the text field.
         */
        function focusHandler(event) {

            switch (event.type) {
            case 'group:focus':
                // save current value, if this group receives initial focus
                initialText = inputState.text = inputNode.val();
                // wait until the <input> element gets focused (see below)
                groupFocused = false;
                break;

            case 'beforedeactivate':
                // Bug 27912: IE keeps text selection visible when losing focus (text
                // is still highlighted although the control is not focused anymore).
                // The 'beforedeactivate' event is triggered by IE only, before the
                // focus leaves the text field. Do not interfere with focus handling
                // in 'blur' events, e.g. Chrome gets confused when changing browser
                // text selection while it currently changes the focus node.
                if (_.browser.IE) {
                    saveInputState();
                    Forms.setInputSelection(inputNode, 0);
                }
                break;

            case 'blur':
                // save field state to be able to restore selection (but not in IE where the
                // selection has been cleared, see the 'beforedeactivate' event handler above)
                if (!_.browser.IE) {
                    saveInputState();
                }
                break;

            case 'group:blur':
                // Bug 27175: always commit value when losing focus
                if (_.isString(initialText) && (initialText !== inputNode.val())) {
                    // pass preserveFocus option to not interfere with current focus handling
                    self.triggerChange(self.getFieldValue(), { sourceEvent: event, preserveFocus: true });
                }
                initialText = null;
                groupFocused = false;
                break;
            }
        }

        /**
         * Handles global 'change:focus' events triggered by the Utils module.
         * Updates the internal state of this instance after the <input> DOM
         * element has received or lost the browser focus. Selects the entire
         * text according to the focus cause (mouse versus keyboard).
         */
        function globalFocusHandler(event, focusNode, blurNode, keyEvent) {

            // the <input> element has received browser focus
            if (inputNode.is(focusNode)) {

                // <input> element focused again (e.g. from drop-down menu): restore selection
                if (groupFocused) {
                    restoreInputState();
                    return;
                }

                // on first focus for the <input> element: remember that this entire
                // group is focused, until the 'group:blur' event will be triggered
                groupFocused = true;

                // remove previous 'mouseup' handler (see below)
                inputNode.off('mouseup');

                // always select entire text when reaching the field with keyboard
                if (_.isObject(keyEvent)) {
                    inputNode.select();
                } else if (select) {
                    // select entire text on mouse click when specified via option
                    inputNode.select();
                    // suppress next 'mouseup' event to prevent that the browser changes the selection
                    inputNode.one('mouseup', false);
                }

                // save current contents and selection
                saveInputState();
                return;
            }

            // the <input> element has lost browser focus
            if (inputNode.is(blurNode)) {
                // always remove the dummy 'mouseup' handler when losing focus (see above)
                inputNode.off('mouseup');
            }
        }

        /**
         * Handles enable/disable events of the group.
         */
        function enableHandler(event, state) {
            if (state) {
                inputNode.removeAttr('readonly');
            } else {
                inputNode.attr('readonly', 'readonly');
            }
        }

        /**
         * Handles 'group:change' events triggered by any action, also by
         * additional controls from sub classes which are not under control of
         * this instance. Prevents triggering a second change event caused by
         * the 'group:blur' event.
         */
        function changeHandler() {
            // prevent another trigger from group:blur handler
            initialText = null;
        }

        /**
         * Handles keyboard events.
         */
        function keyHandler(event) {

            switch (event.keyCode) {
            case KeyCodes.ENTER:
                if (event.type === 'keyup') {
                    self.triggerChange(self.getFieldValue(), { sourceEvent: event });
                }
                return false;
            case KeyCodes.ESCAPE:
                if (event.type === 'keydown') {
                    inputNode.val(initialText);
                    inputHandler();
                }
                break;
            case KeyCodes.HOME:
            case KeyCodes.END:
                // do not let these key codes bubble up, just do the default action
                event.stopPropagation();
                break;
            }
        }

        /**
         * Handles input events triggered when the text changes while typing.
         * Performs live validation with the current validator.
         */
        function inputHandler() {

            var // current text in the input element
                currText = inputNode.val(),
                // previous text contents of the input element
                prevText = inputState.text,
                // result of the text field validation
                result = null;

            // do not perform validation if nothing has changed
            if (currText === prevText) { return; }

            // validate the current field text
            result = validator.validate(currText);

            // update the text field according to the validation result
            if (result === false) {
                // false: restore the old field state
                restoreInputState();
            } else if (_.isString(result) && (result !== currText)) {
                // insert the validation result and restore the old selection
                inputState.text = result;
                restoreInputState();
            }

            // trigger 'group:validate' event to listeners, pass old text
            self.trigger('group:validate', prevText);

            // update current state of the text field
            saveInputState();
        }

        /**
         * Activates/Deactivates a smaller version of the textfield
         *
         * @param {Boolean} value
         *  decides wether the textfield should be small or wide
         */
        function smallTextfield(value){
            // show the small version
            if (value === true) {
                // set new css if exists
                if (_.isObject(smallerVersion.css)) {
                    if (_.isEmpty(savedStyles.css)) { savedStyles.css = wrapperNode.attr('style'); }
                    wrapperNode.css(smallerVersion.css);
                }

                // hide complete label
                if (smallerVersion.hide) { self.getNode().addClass('hidden'); }

            // show the default
            } else {
                // re-set the old (saved) styles
                if (_.has(savedStyles, 'css') && !_.isNull(savedStyles.css)) {
                    wrapperNode.removeAttr('style');
                    wrapperNode.attr('style',savedStyles.css);
                    savedStyles.css = null;
                }

                // hide complete label
                if (smallerVersion.hide) { self.getNode().removeClass('hidden'); }
            }
        }

        // methods ------------------------------------------------------------

        /**
         * Returns the text control element, as jQuery object.
         */
        this.getInputNode = function () {
            return inputNode;
        };

        /**
         * Returns the effective value represented by the current text in the
         * text input control.
         *
         * @returns {Any}
         *  The value represented by the current text, as returned by the
         *  validator of this control group.
         */
        this.getFieldValue = function () {
            return validator.textToValue(inputNode.val());
        };

        /**
         * Sets the effective value to be displayed in the text input control.
         *
         * @param {Any} value
         *  The value to be displayed. The input control will display the text
         *  for this value as returned by the validator of this control group.
         *
         * @returns {TextField}
         *  A reference to this instance.
         */
        this.setFieldValue = function (value) {
            inputNode.val(validator.valueToText(value));
            return this;
        };

        /**
         * Is applying a new validator 
         */
        this.setValidator = function(value) {
            validator = value;
            return this;
        };
        
        /**
         * Converts the passed value to a text using the validator of this
         * control group.
         */
        this.valueToText = function (value) {
            return validator.valueToText(value);
        };

        /**
         * Converts the passed text to a value using the validator of this
         * control group.
         */
        this.textToValue = function (text) {
            return validator.textToValue(text);
        };

        /**
         * Sets the browser focus into the text input element.
         *
         * @returns {TextField}
         *  A reference to this instance.
         */
        this.grabFocus = function () {
            inputNode.focus();
            return this;
        };

        /**
         * Overwrites the base-methods (group.js) to
         * activate/deactivate the small version of the button
         */
        this.activateSmallVersion = function(){
            if (_.isObject(smallerVersion)) {
                smallTextfield(true);
            }
        };
        this.deactivateSmallVersion = function(){
            if (_.isObject(smallerVersion)) {
                smallTextfield(false);
            }
        };

        // initialization -----------------------------------------------------

        // add special marker class used to adjust formatting
        this.getNode().addClass('text-field');

        // insert the text field into this group
        this.addChildNodes(wrapperNode);

        // update text contents in the input field
        this.registerUpdateHandler(updateHandler);

        // register event handlers
        this.on({
            'group:focus group:blur': focusHandler,
            'group:enable': enableHandler,
            'group:change': changeHandler
        });

        inputNode.on({
            'focus beforedeactivate blur': focusHandler,
            'keydown keypress keyup': keyHandler,
            input: inputHandler
        });

        // listen to global 'change:focus' events, to be able to distinguish between mouse and keyboard focus
        this.listenTo(Utils, 'change:focus', globalFocusHandler);

        // forward clicks on the wrapper node (left/right padding beside the input field)
        wrapperNode.on('click', function (event) {
            if (self.isEnabled() && wrapperNode.is(event.target)) {
                inputNode.focus();
            }
        });

        // initialize field state
        saveInputState();

        // destroy all class members on destruction
        this.registerDestructor(function () {
            initOptions = self = inputNode = wrapperNode = validator = null;
        });

    } // class TextField

    // class TextField.Validator ==============================================

    /**
     * Base class for text field validators used to convert between values and
     * field texts, and to validate the text field while editing. Provides a
     * default implementation of all methods that do not restrict editing.
     *
     * @constructor
     *
     * @extends BaseObject
     */
    TextField.Validator = BaseObject.extend({ constructor: function () {

        // methods ------------------------------------------------------------

        /**
         * Converts the passed value to a text string to be inserted into a
         * text field. Intended to be overwritten by derived classes. This
         * default implementation returns the passed value, if it is a string,
         * otherwise an empty string.
         *
         * @param {Any} value
         *  The value to be converted to a text.
         *
         * @returns {String}
         *  The text converted from the passed value, or an empty string, if
         *  the passed value cannot be converted to text.
         *
         */
        this.valueToText = function (value) {
            return _.isString(value) ? value : '';
        };

        /**
         * Converts the passed text to a value that will be passed to all event
         * listeners of a text field. Intended to be overwritten by derived
         * classes. This default implementation returns the unmodified text.
         *
         * @param {String} text
         *  The text to be converted to a value.
         *
         * @returns {Any}
         *  The value converted from the passed text. The value null indicates
         *  that the text cannot be converted to a valid value.
         */
        this.textToValue = function (text) {
            return text;
        };

        /**
         * Validates the passed text that has been changed while editing a text
         * field. It is possible to return a new string value that will be
         * inserted into the text field, or a boolean value indicating whether
         * to restore the old state of the text field. Intended to be
         * overwritten by derived classes. This default implementation does not
         * change the text field.
         *
         * @param {String} text
         *  The current contents of the text field to be validated.
         *
         * @returns {String|Boolean}
         *  When returning a string, the text field will be updated to contain
         *  the returned value while restoring its selection (browsers may
         *  destroy the selection when changing the text). When returning the
         *  boolean value false, the previous state of the text field (as it
         *  was after the last validation) will be restored. Otherwise, the
         *  value of the text field is considered to be valid and will not be
         *  modified.
         */
        this.validate = $.noop;

    }}); // class TextField.Validator

    // global instance of the default validator
    defaultValidator = new TextField.Validator();

    // class TextField.TextValidator ==========================================

    /**
     * A validator for text fields that restricts the allowed values according
     * to the passed options.
     *
     * @constructor
     *
     * @extends TextField.Validator
     *
     * @param {Object} [initOptions]
     *  Optional parameters:
     *  @param {Number} [initOptions.maxLength=0x7FFFFFFF]
     *      The maximum number of characters to be inserted into the text
     *      field. All attempts to insert more characters will be rejected.
     */
    TextField.TextValidator = TextField.Validator.extend({ constructor: function (initOptions) {

        var // maximum length
            maxLength = Utils.getIntegerOption(initOptions, 'maxLength', 0x7FFFFFFF, 0, 0x7FFFFFFF);

        // base constructor ---------------------------------------------------

        TextField.Validator.call(this);

        // methods ------------------------------------------------------------

        this.valueToText = function (value) {
            return _.isString(value) ? value.substr(0, maxLength) : '';
        };

        this.validate = function (text) {
            return text.length <= maxLength;
        };

    }}); // class TextField.TextValidator

    // class TextField.NumberValidator ========================================

    /**
     * A validator for text fields that restricts the allowed values to
     * floating-point numbers.
     *
     * @constructor
     *
     * @extends TextField.Validator
     *
     * @param {Object} [initOptions]
     *  Optional parameters:
     *  @param {Number} [initOptions.min]
     *      The minimum value allowed to enter. If omitted, defaults to
     *      -Number.MAX_VALUE.
     *  @param {Number} [initOptions.max]
     *      The maximum value allowed to enter. If omitted, defaults to
     *      Number.MAX_VALUE.
     *  @param {Number} [initOptions.precision=1]
     *      The precision used to round the current number. The number will be
     *      rounded to multiples of this value. If omitted, the default
     *      precision 1 will round to integers. Must be positive.
     */
    TextField.NumberValidator = TextField.Validator.extend({ constructor: function (initOptions) {

        var // minimum, maximum, and precision
            min = Utils.getNumberOption(initOptions, 'min', -Number.MAX_VALUE, -Number.MAX_VALUE, Number.MAX_VALUE),
            max = Utils.getNumberOption(initOptions, 'max', Number.MAX_VALUE, min, Number.MAX_VALUE),
            precision = Utils.getNumberOption(initOptions, 'precision', 1, 0),
            regex = new RegExp('^' + ((min < 0) ? '-?' : '') + '[0-9]*' + ((precision !== Math.round(precision)) ? '([.,][0-9]*)?' : '') + '$');

        // base constructor ---------------------------------------------------

        TextField.Validator.call(this);

        // methods ------------------------------------------------------------

        this.valueToText = function (value) {
            // TODO L10N of decimal separator
            return _.isFinite(value) ? String(Utils.round(value, precision)) : '';
        };

        this.textToValue = function (text) {
            var value = parseFloat(text.replace(/,/g, '.'));
            return (_.isFinite(value) && (min <= value) && (value <= max)) ? Utils.round(value, precision) : null;
        };

        this.validate = function (text) {
            return regex.test(text);
        };

        this.getMin = function () {
            return min;
        };

        this.getMax = function () {
            return max;
        };

        this.getPrecision = function () {
            return precision;
        };

        this.restrictValue = function (value) {
            return _.isFinite(value) ? Utils.minMax(value, min, max) : value;
        };

    }}); // class TextField.NumberValidator

    // exports ================================================================

    // derive this class from class Group
    return Group.extend({ constructor: TextField });

});
