/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/tk/logger', ['io.ox/office/tk/config'], function (Config) {

    'use strict';

    var // global indent for all loggers
        indent = '',

        // prevent collapsing space characters in browser console
        INDENT_PARTICLE = '\xa0 ';

    // class Logger ===========================================================

    /**
     * Provides methods to create logging messages in the browser console.
     * Logging will always be disabled without global debug mode specified in
     * the server-side configuration option 'debugavailable'. If the
     * application URL contains the anchor option 'office:log-timestamp=true',
     * all log message will be prepended with the current system time.
     *
     * @constructor
     *
     * @param {Object} [initOptions]
     *  Additional optional parameters:
     *  @param {Boolean|String} [initOptions.enable=true]
     *      Specifies whether the logger is enabled. If set to a string, the
     *      application URL will be checked for an according anchor option set
     *      to the value 'true'.
     *  @param {String} [initOptions.prefix]
     *      If specified, all messages will be prefixed with the passed text
     *      (enclosed in bracket characters).
     */
    function Logger(initOptions) {

        var // the 'enable' option passed to the constructor (defaults to true)
            enableOpt = (_.isObject(initOptions) && ('enable' in initOptions)) ? initOptions.enable : true,

            // whether the logger is enabled
            enabled = Config.DEBUG && (_.isString(enableOpt) ? Config.getUrlFlag(enableOpt) : (_.isBoolean(enableOpt) && enableOpt)),

            // the prefix printed before each message
            prefix = (_.isObject(initOptions) && _.isString(initOptions.prefix)) ? ('[' + initOptions.prefix  + ']') : null;

        // private methods ----------------------------------------------------

        /**
         * Writes a message to the browser output console.
         *
         * @param {String} level
         *  The log level of the message. The string 'log' will create a
         *  generic log message, the string 'info' will create an information,
         *  the string 'warn' will create a warning message, and the string
         *  'error' will create an error message.
         */
        function log(level, args) {

            // IE9: console missing if debugger is not running
            if (!_.isObject(window.console)) { return; }

            // add white-space indent
            if (indent.length > 0) {
                args.unshift(indent);
            }

            // add logger prefix
            if (prefix) {
                args.unshift(prefix);
            }

            // check that the browser console supports the operation
            if (!_.isFunction(window.console[level])) {
                args.unshift(level.toUpperCase());
                level = 'log';
            }

            // add time stamp
            if (Logger.LOG_TIMESTAMP) {
                var time = new Date();
                args.unshift(time.toLocaleTimeString() + '.' + ('00' + (time.getTime() % 1000)).slice(-3));
            }

            if (_.isFunction(window.console[level])) {
                window.console[level].apply(window.console, args);
            } else if (_.isNumber(_.browser.IE) && (_.browser.IE < 10)) {
                window.console.log(args.join(' '));
            }
        }

        // methods ------------------------------------------------------------

        /**
         * Returns whether this logger is active.
         *
         * @returns {Boolean}
         *  Whether this logger is active.
         */
        this.isLoggingActive = function () {
            return enabled;
        };

        /**
         * Invokes the passed callback function if and only if this logger is
         * enabled. Can be used to prevent running rather complex calculations
         * that are needed for debug logging only.
         *
         * @param {Function} callback
         *  The callback function.
         *
         * @param {Object} [context]
         *  The context to be bound to the callback function.
         *
         * @returns {Logger}
         *  A reference to this instance.
         */
        this.withLogging = function (callback, context) {
            if (enabled) { callback.call(context); }
            return this;
        };

        /**
         * Writes a log message to the browser output console, if debug mode is
         * enabled in the global configuration.
         *
         * @param {Any} [...]
         *  The values to be written to the console.
         */
        this.log = enabled ? function () { log('log', _.toArray(arguments)); } : $.noop;

        /**
         * Writes an info message to the browser output console, if debug mode is
         * enabled in the global configuration.
         *
         * @param {Any} [...]
         *  The values to be written to the console.
         */
        this.info = enabled ? function () { log('info', _.toArray(arguments)); } : $.noop;

        /**
         * Writes a warning message to the browser output console, if debug mode is
         * enabled in the global configuration.
         *
         * @param {Any} [...]
         *  The values to be written to the console.
         */
        this.warn = enabled ? function () { log('warn', _.toArray(arguments)); } : $.noop;

        /**
         * Writes an error message to the browser output console, if debug mode is
         * enabled in the global configuration.
         *
         * @param {Any} [...]
         *  The values to be written to the console.
         */
        this.error = enabled ? function () { log('error', _.toArray(arguments)); } : $.noop;

        /**
         * Writes an error message to the browser output console describing the
         * passed exception object, if debug mode is enabled in the global
         * configuration.
         *
         * @param {Any} exception
         *  The exception as caught in a try/catch.
         *
         * @param {Any} [...]
         *  Additional values to be written to the console.
         */
        this.exception = enabled ? function (exception) {

            var // the stack contained in the exception object
                stack = _.isObject(exception) && (exception.stack || exception.stacktrace);

            log('error', ['Exception caught:', exception].concat(_.toArray(arguments).slice(1)));
            if (stack) { log('error', ['Stack trace:', stack]); }

        } : $.noop;

        /**
         * Registers an event handler for the specified event and writes a log
         * message whenever the event has been triggered.
         *
         * @param {Object} source
         *  The event source object.
         *
         * @param {String} type
         *  The name of the event. Can be a space-separated list.
         *
         * @param {String} [...]
         *  The names for additional parameters passed to the event handler.
         */
        this.logEvent = enabled ? function (source, type) {

            var // self reference
                logFunc = _.bind(function (args) { this.log.apply(this, args); }, this),
                // the event parameter names
                paramNames = _.toArray(arguments).slice(2),
                // if the source object provides a DOM node, write it to the log message too
                rootNode = _.isFunction(source.getNode) ? $(source.getNode())[0] : null;

            source.on(type, function (event) {

                var // all arguments passed to the log message
                    logArgs = ['type="' + event.type + '"', 'this=', this];

                // add the root node of the event source
                if (rootNode) { logArgs.push('node=', rootNode); }

                // add the additional event parameters
                _.chain(arguments).toArray().slice(1).each(function (arg, index) {
                    if (index < paramNames.length) { logArgs.push(paramNames[index] + '='); }
                    logArgs.push(arg);
                });

                logFunc(logArgs);
            });
        } : $.noop;

        /**
         * Executes the passed callback function and writes a message to the
         * browser console containing the execution time of the callback. The
         * callback function will be invoked ALWAYS, regardless whether logging
         * is enabled or not.
         *
         * @param {String} message
         *  The initial message printed to the browser console before invoking
         *  the callback function. The execution time will be shown after the
         *  callback function has returned.
         *
         * @param {Function} callback
         *  The callback function to be executed.
         *
         * @param {Object} [context]
         *  The context to be bound to the callback function.
         *
         * @returns {Any}
         *  The return value of the callback function.
         */
        this.takeTime = enabled ? (function (logger) {

            var now = (window.performance && window.performance.now) ? _.bind(window.performance.now, window.performance) : _.now;

            return function (message, callback, context) {
                var t0 = 0;
                try {
                    logger.log('=>', message);
                    t0 = now();
                    return Logger.indent(callback, context);
                } finally {
                    logger.log('<=', (Math.round((now() - t0) * 1000) / 1000) + 'ms');
                }
            };
        })(this) : function (message, callback, context) {
            return callback.call(context);
        };

        /**
         * Creates and returns a version of the passed callback function that
         * shows the execution time of each its invocation, if this logger is
         * enabled (wraps the callback function in a call to Logger.takeTime()
         * internally).
         *
         * @param {String} message
         *  The initial message printed to the browser console before invoking
         *  the callback function. The execution time will be shown after the
         *  callback function has returned.
         *
         * @param {Function} method
         *  The callback function to be profiled.
         *
         * @returns {Function}
         *  The passed callback function wrapped into a call to the method
         *  Logger.takeTime(). Forwards the current calling context to the
         *  callback function, and returns its return value.
         */
        this.profileMethod = enabled ? function (message, method) {
            var logger = this;
            return function () {
                var args = arguments;
                return logger.takeTime(message, function () {
                    return method.apply(this, args);
                }, this);
            };
        } : function (message, method) {
            return method;
        };

    } // class Logger

    // static methods ---------------------------------------------------------

    /**
     * Extends the passed object or class with logging functionality.
     *
     * @returns {Object|Function}
     *  An object that will be extended with the logging methods provided by
     *  the Logger class. Can be a class constructor function, in this case the
     *  class will contain static logging methods.
     *
     * @returns {Object|Function}
     *  A reference to the passed and extended object or class.
     */
    Logger.extend = function (obj) {
        Logger.apply(obj, _.toArray(arguments).slice(1));
        return obj;
    };

    /**
     * Invokes the passed callback function with increased indentation for all
     * active loggers.
     *
     * @param {Function} callback
     *  The callback function to be executed.
     *
     * @param {Object} [context]
     *  The context to be bound to the callback function.
     *
     * @returns {Any}
     *  The return value of the callback function.
     */
    Logger.indent = function (callback, context) {
        try {
            indent += INDENT_PARTICLE;
            return callback.call(context);
        } finally {
            indent = indent.substring(INDENT_PARTICLE.length);
        }
    };

    /**
     * Forwards console output into a fixed DOM node. Especially useful for
     * mobile devices and a better stability (some browsers, e.g. Chrome, have
     * problems with very large console logs).
     */
    Logger.enableDOMConsole = Config.DEBUG ? _.once(function () {

        var ICONS = { info: 'fa-info-circle', warn: 'fa-warning', error: 'fa-times-circle' },
            consoleNode = $('<div>', { id: 'io-ox-office-console' }).addClass('noI18n collapsed').appendTo('body'),
            outputNode = $('<div>').addClass('output').appendTo(consoleNode),
            buttonsNode = $('<tr>').appendTo($('<table>').addClass('buttons').appendTo(consoleNode)),
            lockButton = $('<td>').text(_.noI18n('Scroll Lock')).appendTo(buttonsNode),
            selectButton = $('<td>').text(_.noI18n('Select All')).appendTo(buttonsNode),
            clearButton = $('<td>').text(_.noI18n('Clear')).appendTo(buttonsNode),
            hideButton = $('<td>').text(_.noI18n('Hide')).appendTo(buttonsNode),
            locked = false;

        function writeToDOMConsole(methodName, args) {
            var rowNode = $('<p>').addClass(methodName).appendTo(outputNode);
            if (Logger.LOG_TIMESTAMP) { rowNode.append($('<span>').text(args.shift())); }
            if (methodName in ICONS) { rowNode.append('<i class="fa ' + ICONS[methodName] + '"></i>'); }
            _.each(args, function (msg) { rowNode.append($('<span>').text(msg)); });
            if (!locked) { outputNode.scrollTop(outputNode[0].scrollHeight); }
        }

        consoleNode.append('<i class="fa fa-bug"></i>');

        // IE9: console missing if debugger is not running
        if (!_.isObject(window.console)) { window.console = {}; }

        // prepare button click handlers
        lockButton.on('click', function () {
            locked = !locked;
            lockButton.toggleClass('selected', locked);
            return false;
        });
        selectButton.on('click', function () {
            var docRange = window.document.createRange();
            docRange.setStart(consoleNode[0], 0);
            docRange.setEnd(consoleNode[0], 1);
            window.getSelection().removeAllRanges();
            window.getSelection().addRange(docRange);
            return false;
        });
        clearButton.on('click', function () {
            outputNode.empty();
            return false;
        });
        hideButton.on('click', function () {
            consoleNode.addClass('collapsed');
            return false;
        });
        consoleNode.on('click', function () {
            if (consoleNode.hasClass('collapsed')) { consoleNode.removeClass('collapsed'); }
            return false;
        });

        // replace implementations of browser console methods
        _.each(['log', 'info', 'warn', 'error'], function (methodName) {
            var origMethod = window.console[methodName];
            window.console[methodName] = function () {
                writeToDOMConsole(methodName, _.toArray(arguments));
                return _.isFunction(origMethod) ? origMethod.apply(this, arguments) : undefined;
            };
        });

        // global exception handler to forward unhandled exceptions to the DOM console
        if (_.isFunction(window.onerror)) {
            window.console.warn('Logger.enableDOMConsole(): global error handler exists already');
        } else {
            window.onerror = function () {
                writeToDOMConsole('error', _.toArray(arguments));
                return false;
            };
        }
    }) : $.noop;

    // constants --------------------------------------------------------------

    // whether to add a time stamp to all messages
    Logger.LOG_TIMESTAMP = Config.DEBUG && Config.getUrlFlag('office:log-timestamp');

    // global initialization ==================================================

    // forward console output into a fixed DOM node
    if (Config.DEBUG && Config.getUrlFlag('office:debug-console')) {
        Logger.enableDOMConsole();
    }

    // exports ================================================================

    return Logger;

});
