/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/tk/popup/basemenu',
    ['io.ox/office/tk/config',
     'io.ox/office/tk/utils',
     'io.ox/office/tk/keycodes',
     'io.ox/office/tk/forms',
     'io.ox/office/tk/popup/basepopup'
    ], function (Config, Utils, KeyCodes, Forms, BasePopup) {

    'use strict';

    // class BaseMenu =========================================================

    /**
     * Wrapper class for a DOM node used as interactive pop-up menu element,
     * shown on top of the application window, and relative to an arbitrary DOM
     * node. The class adds basic browser focus handling. On opening, the
     * browser focus will be moved invisibly into the menu element to enable
     * keyboard events (focus navigation) inside the menu. When the browser
     * focus leaves the menu node (regardless how that happened, e.g. clicks or
     * taps anywhere outside the menu, or special key shortcuts to move the
     * focus away), the menu will close itself automatically.
     *
     * @constructor
     *
     * @extends BasePopup
     *
     * @param {Object} [initOptions]
     *  Initial options to control the properties of the pop-up menu element.
     *  Supports all options also supported by the base class BasePopup.
     *  Additionally, the following options are supported:
     *  @param {Boolean} [initOptions.autoFocus=true]
     *      If set to true (or omitted), the browser focus will be moved into
     *      the pop-up menu after it has been opened. If set to false, the
     *      browser focus will not be changed. In this case, the node currently
     *      focused must be contained in the option 'focusableNodes' to prevent
     *      that the menu will be closed immediately afterwards due to missing
     *      menu focus.
     *  @param {Boolean} [initOptions.autoClose=true]
     *      If set to true (or omitted), and the browser focus leaves the
     *      pop-up menu node (due to outside clicks, TAB key, etc.), the menu
     *      will be closed automatically.
     *  @param {HTMLElement|jQuery|Function} [initOptions.focusableNodes]
     *      Additional DOM nodes located outside the root node of this pop-up
     *      menu that do not cause to close the menu automatically while they
     *      are focused. More focusable nodes can be registered at runtime with
     *      the method BaseMenu.registerFocusableNodes(). Can be a function
     *      that will be called on every focus change, and that must return
     *      a jQuery object or a plain DOM element.
     *  @param {String|Function} [initOptions.preferFocusFilter]
     *      A jQuery selector that filters the available focusable control
     *      elements to a list of preferred controls, that will be used when
     *      moving the browser focus into this pop-up menu using the method
     *      BaseMenu.grabFocus(). The selector will be passed to the jQuery
     *      method jQuery.filter(). If this selector is a function, it will be
     *      called with the DOM node bound to the symbol 'this'. See the jQuery
     *      API documentation at http://api.jquery.com/filter for details.
     */
    function BaseMenu(initOptions) {

        var // self reference
            self = this,

            // whether to focus the menu after it has been opened
            autoFocus = Utils.getBooleanOption(initOptions, 'autoFocus', true),

            // whether to close the menu automatically after losing focus
            autoClose = Utils.getBooleanOption(initOptions, 'autoClose', true),

            // static collection of DOM elements preventing auto-close of the menu
            focusableNodes = $(),

            // dynamic callbacks for additional focusable DOM elements preventing auto-close
            focusableNodeResolvers = [],

            // filter selector for preferred focusable elements
            preferFocusFilter = Utils.getOption(initOptions, 'preferFocusFilter'),

            // events registered at the document while the menu is open
            DOCUMENT_EVENTS = 'mousedown touchstart';

        // base constructor ---------------------------------------------------

        BasePopup.call(this, initOptions);

        // private methods ----------------------------------------------------

        /**
         * Returns whether the passed DOM node is contained in the list of
         * focusable nodes which prevent auto-closing the menu if it loses the
         * browser focus.
         */
        function isFocusableNode(node) {

            // bug 33548: treat IE's focus helper node as focusable node (see Utils.clearBrowserSelection() for details)
            if ($(node).is('.focus-helper')) { return true; }

            var // all focusable nodes (start with static nodes)
                allFocusableNodes = focusableNodes;

            // add dynamic focusable nodes
            _.each(focusableNodeResolvers, function (resolver) {
                allFocusableNodes = allFocusableNodes.add(resolver.call(self));
            });

            // check whether the passed node is contained in the focusable nodes (directly, or as descendant)
            return (allFocusableNodes.filter(node).length > 0) || (allFocusableNodes.has(node).length > 0);
        }

        /**
         * Closes this menu if the passed node is not included in the
         * registered focusable nodes.
         */
        function hideMenuOnFocusNode(node) {
            if (autoClose && !(node && isFocusableNode(node))) {
                // do not close the menu, if the debug URL flag 'office:sticky-popups' is set (but still return true)
                if (!Config.DEBUG || !Config.getUrlFlag('office:sticky-popups')) {
                    self.hide();
                }
                return true;
            }
            return false;
        }

        /**
         * Hides the menu node automatically when the browser focus leaves.
         */
        function changeFocusHandler(event, focusNode) {

            // close the menu unless the focus has been moved to the registered nodes
            if (hideMenuOnFocusNode(focusNode)) {
                return;
            }

            // scroll the pop-up menu to make the focused node visible
            if (focusNode && Utils.containsNode(self.getContentNode(), focusNode)) {
                self.scrollToChildNode(focusNode);
            }
        }

        /**
         * Handles mouse clicks everywhere on the page, and closes the context
         * menu automatically.
         */
        function globalClickHandler(event) {
            hideMenuOnFocusNode(event.target);
        }

        /**
         * Initializes the pop-up menu after it has been opened.
         */
        function popupShowHandler() {

            // move focus into the menu node, for keyboard handling
            if (autoFocus) { self.getContentNode().focus(); }

            // listen to focus events, hide the menu when focus leaves the menu
            Utils.on('change:focus', changeFocusHandler);

            // Add global click handler that closes the menu automatically when
            // clicking somewhere in the page without changing the browser focus
            // (elements without tab-index). The event handler must be attached
            // deferred, otherwise it would close the context menu immediately
            // after it has been opened with a mouse click while the event bubbles
            // up to the document root.
            _.defer(function () {
                if (self.isVisible()) {
                    $(document).on(DOCUMENT_EVENTS, globalClickHandler);
                }
            });
        }

        /**
         * Deinitializes the pop-up menu before it will be closed.
         */
        function popupBeforeHideHandler() {
            // unregister the focus handler
            Utils.off('change:focus', changeFocusHandler);
            // unregister the global click/touch handler
            $(document).off(DOCUMENT_EVENTS, globalClickHandler);
        }

        /**
         * Hides the pop-up menu, if the ESCAPE key has been pressed inside.
         */
        function keyDownHandler(event) {
            if (event.keyCode === KeyCodes.ESCAPE) {
                self.hide();
                return false;
            }
        }

        // methods ------------------------------------------------------------

        /**
         * Returns whether this pop-up menu contains the browser focus.
         *
         * @returns {Boolean}
         *  Whether this pop-up menu contains the browser focus.
         */
        this.hasFocus = function () {
            return Utils.containsFocus(this.getNode());
        };

        /**
         * Sets the browser focus into the first focusable control element in
         * this pop-up menu element.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {Boolean} [options.bottom=false]
         *      If set to true, the last available entry of the pop-up menu
         *      will be focused instead of the first.
         *
         * @returns {BaseMenu}
         *  A reference to this instance.
         */
        this.grabFocus = function (options) {

            // do nothing if a descendant of the content node is already focused
            if (Utils.containsFocus(this.getContentNode())) { return this; }

            var // all focusable control elements
                focusNodes = Forms.findFocusableNodes(this.getNode()),
                // the preferred controls returned by the callback function
                preferredNodes = preferFocusFilter ? focusNodes.filter(preferFocusFilter) : focusNodes,
                // whether to select the last control
                bottom = Utils.getBooleanOption(options, 'bottom', false);

            // fall back to all focusable controls if no preferred controls are available
            if (preferredNodes.length === 0) { preferredNodes = focusNodes; }
            preferredNodes[bottom ? 'last' : 'first']().focus();
            return this;
        };

        /**
         * Registers one or more DOM nodes located outside the root node of
         * this pop-up menu that do not cause to close the menu automatically
         * while they are focused.
         *
         * @param {HTMLElement|jQuery|Function} nodes
         *  The DOM node(s) to be added to the list of focusable elements. Can
         *  be a function that will be called on every focus change, and that
         *  must return a jQuery object or a plain DOM element.
         *
         * @returns {BaseMenu}
         *  A reference to this instance.
         */
        this.registerFocusableNodes = function (nodes) {
            if ((nodes instanceof HTMLElement) || (nodes instanceof $)) {
                focusableNodes = focusableNodes.add(nodes);
            } else if (_.isFunction(nodes)) {
                focusableNodeResolvers.push(nodes);
            }
            return this;
        };

        // initialization -----------------------------------------------------

        // add basic pop-up menu styling
        this.getNode().addClass('popup-menu');

        // initialize static focusable nodes and dynamic callbacks for focusable nodes
        this.registerFocusableNodes(Utils.getOption(initOptions, 'focusableNodes'));

        // make the content node focusable to be able to provide keyboard support
        // without actually focusing a visible node inside the menu
        this.getContentNode().attr('tabindex', -1);
        this.registerFocusableNodes(this.getNode());

        // register event handlers
        this.on({ 'popup:show': popupShowHandler, 'popup:beforehide': popupBeforeHideHandler });
        this.getNode().on('keydown', keyDownHandler);

        // Bug 35239: Windows/Chrome only: When clicking a button in a pop-up menu while
        // the browser tool-tip of the button is visible, the browser will not generate
        // a 'click' event (every second time only). Workaround is to prevent the default
        // action of the 'mousedown' events on button elements.
        if (_.browser.Windows && _.browser.Chrome) {
            this.getNode().on('mousedown', Forms.BUTTON_SELECTOR, function (evt) { evt.preventDefault(); });
        }

        // destroy all class members on destruction
        this.registerDestructor(function () {
            initOptions = self = focusableNodes = focusableNodeResolvers = preferFocusFilter = null;
        });

    } // class BaseMenu

    // exports ================================================================

    // derive this class from class BasePopup
    return BasePopup.extend({ constructor: BaseMenu });

});
