/**
 * All content on this website (including text, images, source
 * code and any other original works), unless otherwise noted,
 * is licensed under a Creative Commons License.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) Open-Xchange Inc., 2006-2012
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/tk/utils',
    ['io.ox/core/event',
     'io.ox/office/tk/config',
     'io.ox/office/tk/keycodes',
     'io.ox/office/tk/logger',
     'less!io.ox/office/tk/style'
    ], function (Events, Config, KeyCodes, Logger) {

    'use strict';

    var // the global element used to store DOM elements without showing them
        hiddenStorageNode = $('<div>', { id: 'io-ox-office-temp' }).prependTo('body'),

        // a helper node to workaround selection problems in IE (bug 28515, bug 28711)
        focusHelperNode = $('<div class="focus-helper" contenteditable="true">'),

        // tabindex value, if the element should removed from the tab-flow
        untabableTabindex = '-9',

        // tabindex value, if the element should add to the tab-flow
        tabableTabindex = '1',

        // a global timer that can be used for performance debugging
        globalTimer = null,

        // a global counter that can be used for performance debugging
        globalCounter = null,

        // whether the times of several following global timers shall be accumulated
        accumulateTime = false,

        // the accumulated time of several following global timers
        accumulatedTime = null,

        // the number of accumulated values
        accumulatedCounter = null;

    // static class Utils =====================================================

    var // the exported Utils class
        Utils = {};

    // static extensions ------------------------------------------------------

    Events.extend(Utils);
    Logger.extend(Utils);

    // constants --------------------------------------------------------------

    /**
     * A unique object used as return value in callback functions of iteration
     * loops to break the iteration process immediately.
     *
     * @constant
     */
    Utils.BREAK = {};

    /**
     * Number of colors in the color scheme used for highlighting of specific
     * document contents.
     *
     * @constant
     */
    Utils.SCHEME_COLOR_COUNT = 7;

    /**
     * A Boolean flag specifying whether the current display is a retina
     * display.
     *
     * @constant
     */
    Utils.RETINA = _.device('retina');

    /**
     * A Boolean flag specifying whether the current device is small (smart
     * phones).
     *
     * @constant
     */
    Utils.SMALL_DEVICE = _.device('smartphone');

    /**
     * A Boolean flag specifying whether the current device is small (smart
     * phones) or medium sized (tablets with a width of less than 1024px).
     *
     * @constant
     */
    Utils.COMPACT_DEVICE = Utils.SMALL_DEVICE || _.device('tablet');

    /**
     * A Boolean flag specifying whether the Internet Explorer version 9 is
     * running.
     *
     * @constant
     */
    Utils.IE9 = _.isNumber(_.browser.IE) && (_.browser.IE < 10);

    /**
     * A Boolean flag specifying whether the current device is an iPad.
     *
     * @constant
     */
    Utils.IPAD = !!(Modernizr.touch && _.browser.iOS && _.browser.Safari);

    /**
     * A Boolean flag specifying whether the current device is Android and the
     * browser is Chrome.
     *
     * @constant
     */
     Utils.CHROME_ON_ANDROID = !!(_.browser.Android && _.browser.Chrome);

    /**
     * Determines the performance level of the system using some system and
     * browser specific values.
     *
     * @returns {Number}
     *  A performance level, as percentage between 0 and 100, where the value 0
     *  means poor performance, and the value 100 best performance.
     */
    Utils.PERFORMANCE_LEVEL = (function () {

        var // default factor for unknown devices/browsers
            DEFAULT_FACTOR = 0.8,
            // the resulting factor
            resultLevel = 100;

        function getAspectFactor(aspects) {
            var aspectFactor = _.find(aspects, function (factor, key) { return _.device(key); });
            return _.isNumber(aspectFactor) ? aspectFactor : DEFAULT_FACTOR;
        }

        // reduce performance level by device type
        resultLevel *= getAspectFactor({ desktop: 1, tablet: 0.6, smartphone: 0.5 });
        // reduce performance level by browser type
        resultLevel *= getAspectFactor({ chrome: 1, safari: 1, firefox: 0.9, ie: 0.6 });

        // ensure value is at least 10%
        resultLevel = Math.max(10, resultLevel);
        Utils.log('Utils: platform performance level: ' + (Math.round(resultLevel * 10) / 10) + '%');
        return resultLevel;
    }());

    /**
     * A Boolean flag specifying whether the current device is considered to be
     * slow.
     *
     * @constant
     */
    Utils.SLOW_DEVICE = Utils.PERFORMANCE_LEVEL <= 50;

    /**
     * The maximum explicit size of single DOM nodes, in pixels. The size is
     * limited by browsers; by using larger sizes the elements may collapse to
     * zero size.
     * - IE limits this to 1,533,917 (2^31/1400) in both directions.
     * - Firefox limits this to 17,895,696 (0x1111110) in both directions.
     * - Chrome limits this to 33,554,428 (0x1FFFFFC) in both directions.
     *
     * @constant
     */
    Utils.MAX_NODE_SIZE = _.browser.IE ? 1.5e6 : _.browser.WebKit ? 33.5e6 : 17.8e6;

    /**
     * The maximum width of a container node that can be reached by inserting
     * multiple child nodes, in pixels. The width is limited by browsers; by
     * inserting more nodes the width of the container node becomes incorrect
     * or collapses to zero.
     * - IE limits this to 10,737,418 (2^31/200) pixels.
     * - Firefox and Chrome do not allow to extend the container node beyond
     *   the explicit size limits of a single node (see comment for the
     *   Utils.MAX_NODE_SIZE constant above).
     *
     * @constant
     */
    Utils.MAX_CONTAINER_WIDTH = _.browser.IE ? 10.7e6 : Utils.MAX_NODE_SIZE;

    /**
     * The maximum height of a container node that can be reached by inserting
     * multiple child nodes, in pixels. The height is limited by browsers; by
     * inserting more nodes the height of the container node becomes incorrect
     * or collapses to zero.
     * - IE limits this to 21,474,836 (2^31/100) pixels.
     * - Firefox and Chrome do not allow to extend the container node beyond
     *   the explicit size limits of a single node (see comment for the
     *   Utils.MAX_NODE_SIZE constant above).
     *
     * @constant
     */
    Utils.MAX_CONTAINER_HEIGHT = _.browser.IE ? 21.4e6 : Utils.MAX_NODE_SIZE;

    // browser helpers --------------------------------------------------------

    /**
     * Checks that a specific browser at least has the necessary major version.
     *
     * @param {String}
     *  The name of the browser, e.g. Chrome, Firefox, IE, Safari
     *
     * @param {Number}
     *  The minimal major version of browser.
     *
     * @returns {Boolean}
     *  Returns TRUE if the browser name and the minimal major version matches,
     *  otherwise FALSE.
     */
    Utils.isBrowserAndMinVersion = function(browser, neededMajorVersion) {
        var version = _.browser[browser];

        return _.isNumber(version) && (version > (neededMajorVersion - 1));
    };

    /**
     * Checks that on Android the browser is Chrome and the major version
     * is at least a specific version to to enable text editing.
     *
     * @param {Boolean}
     *  Returns TRUE if the current system is Android and the browser is
     *  Chrome and the major version is at least the necessary version, to
     *  enable editing. It returns FALSE if one of the previous conditions
     *  is not fulfilled.
     */
    Utils.supportedChromeVersionOnAndroidForText = function () {
        // TODO: the major version must be adapted as soon as we know
        // what version will include the necessary fixes to enable
        // editing.
        // (11.08.2014 we think that M38 will include the fixes),
        return _.browser.Android && Utils.isBrowserAndMinVersion('Chrome', 38);
    };

    // generic JS object helpers ----------------------------------------------

    /**
     * Returns a new object containing a single property with the specified key
     * and value.
     *
     * @param {String} key
     *  The name of the property to be inserted into the returned object.
     *
     * @param value
     *  The value of the property to be inserted into the returned object.
     *
     * @returns {Object}
     *  A new object with a single property.
     */
    Utils.makeSimpleObject = function (key, value) {
        var object = {};
        object[key] = value;
        return object;
    };

    /**
     * Compares the two passed numeric arrays lexicographically (starting with
     * the first element of both arrays, and visiting the following elements
     * until a pair of elements is different).
     *
     * @param {Number[]} array1
     *  The first array that will be compared to the second array.
     *
     * @param {Number[]} array2
     *  The second array that will be compared to the first array.
     *
     * @param {Number} [maxLength]
     *  If specified, compares the specified number of leading elements in the
     *  passed arrays. Otherwise, the entire arrays will be compared.
     *
     * returns {Number}
     *  A negative value, if array1 is lexicographically less than array2; a
     *  positive value, if array1 is lexicographically greater than array2; or
     *  zero, if both arrays are equal.
     */
    Utils.compareNumberArrays = function (array1, array2, maxLength) {

        var // number of elements in the arrays to be compared
            length1 = _.isNumber(maxLength) ? Math.min(array1.length, maxLength) : array1.length,
            length2 = _.isNumber(maxLength) ? Math.min(array2.length, maxLength) : array2.length,
            // minimum length of both arrays
            length = Math.min(length1, length2),
            // loop index
            index = 0;

        // compare all array elements
        for (index = 0; index < length; index += 1) {
            if (array1[index] !== array2[index]) {
                return array1[index] - array2[index];
            }
        }

        // all compared elements are equal, compare array lengths (restricted to the passed maximum length)
        return length1 - length2;
    };

    /**
     * Returns whether the passed object contains the specified property.
     *
     * @param {Object} object
     *  The object to be checked for its properties.
     *
     * @param {String|RegExp} propertyName
     *  The name of the property to be checked. If set to a regular expression,
     *  at least one of the object's property names has to match it.
     *
     * @returns {Boolean}
     *  Whether the specified property exists in the object.
     */
    Utils.hasProperty = function (object, propertyName) {
        return _.isString(propertyName) ? _.has(object, propertyName) :
            (_.isRegExp(propertyName) && _.any(object, function (value, name) { return propertyName.test(name); }));
    };

    /**
     * Returns whether the passed object contains any of the specified
     * properties.
     *
     * @param {Object} object
     *  The object to be checked for its properties.
     *
     * @param {String[]} propertyNames
     *  The names of all properties to be checked.
     *
     * @returns {Boolean}
     *  Whether at least one of the specified properties exists in the object.
     */
    Utils.hasAnyProperty = function (object, propertyNames) {
        return _.any(propertyNames, Object.prototype.hasOwnProperty, object);
    };

    /**
     * Returns whether the passed object contains all of the specified
     * properties.
     *
     * @param {Object} object
     *  The object to be checked for its properties.
     *
     * @param {String[]} propertyNames
     *  The names of all properties to be checked.
     *
     * @returns {Boolean}
     *  Whether all of the specified properties exists in the object.
     */
    Utils.hasAllProperties = function (object, propertyNames) {
        return _.all(propertyNames, Object.prototype.hasOwnProperty, object);
    };

    /**
     * Returns whether the specified properties of the passed objects are
     * equal, while ignoring all other properties. If a property is missing in
     * both objects, it is considered to be equal. Uses a custom comparator or
     * the Underscore method _.isEqual() to compare the property values.
     *
     * @param {Object} object1
     *  The first object to be compared to the other.
     *
     * @param {Object} object2
     *  The second object to be compared to the other.
     *
     * @param {String[]} propertyNames
     *  The names of all properties of the objects that will be compared.
     *
     * @param {Function} [comparator=_.isEqual]
     *  A binary predicate function that returns true if the passed property
     *  values are considered being equal. Will be called, if both objects
     *  passed to this method contain a specific property, and receives the
     *  property values from both objects.
     *
     * @returns {Boolean}
     *  Whether all specified properties are equal in both objects.
     */
    Utils.hasEqualProperties = function (object1, object2, propertyNames, comparator) {

        // default to the _isEqual() method to compare property values
        comparator = _.isFunction(comparator) ? comparator : _.isEqual;

        // process all specified properties
        return _.all(propertyNames, function (propName) {
            var hasProp1 = _.has(object1, propName),
                hasProp2 = _.has(object2, propName);
            return (hasProp1 === hasProp2) && (!hasProp1 || comparator(object1[propName], object2[propName]));
        });
    };

    /**
     * Removes all occurrences of the specified value from the array in-place.
     * Uses a custom comparator or the Underscore method _.isEqual() to compare
     * the property values.
     *
     * @param {Array} array
     *  (in/out) The array to be shortened.
     *
     * @param {Any} value
     *  The value to be removed from the array.
     *
     * @param {Function} [comparator=_.isEqual]
     *  A binary predicate function that returns true if the two passed values
     *  are considered being equal. Will be called to compare the passed value
     *  against the elements of the array.
     *
     * @returns {Number}
     *  The number of elements removed from the array.
     */
    Utils.spliceValue = function (array, value, comparator) {

        var // the old array length
            length = array.length;

        // default to the _isEqual() method to compare values
        comparator = _.isFunction(comparator) ? comparator : _.isEqual;

        // remove all elements matching the passed value
        for (var index = length - 1; index >= 0; index -= 1) {
            if (comparator(value, array[index])) {
                array.splice(index, 1);
            }
        }

        return length - array.length;
    };

    /**
     * Calls the passed iterator function for all elements of the passed array.
     *
     * @param {Array|jQuery} array
     *  The array whose elements will be iterated. jQuery collections are
     *  also supported.
     *
     * @param {Function} iterator
     *  The iterator function invoked for all elements in the array. Receives
     *  the following parameters:
     *  (1) {Any} element
     *      The current array element.
     *  (2) {Number} index
     *      The index of the array element.
     *  (3) {Array} array
     *      The array passed to this method.
     *  If the iterator returns the Utils.BREAK object, the iteration process
     *  will be stopped immediately.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Object} [options.context]
     *      If specified, the iterator will be called with this context (the
     *      symbol 'this' will be bound to the context inside the iterator
     *      function).
     *  @param {Boolean} [options.reverse=false]
     *      If set to true, the elements will be visited in reversed order.
     *
     * @returns {Utils.BREAK|Undefined}
     *  A reference to the Utils.BREAK object, if the iterator has returned
     *  Utils.BREAK to stop the iteration process, otherwise undefined.
     */
    Utils.iterateArray = function (array, iterator, options) {

        var context = Utils.getOption(options, 'context');

        if (Utils.getBooleanOption(options, 'reverse', false)) {
            for (var index = array.length - 1; index >= 0; index -= 1) {
                if (iterator.call(context, array[index], index, array) === Utils.BREAK) {
                    return Utils.BREAK;
                }
            }
        } else {
            for (var index = 0; index < array.length; index += 1) {
                if (iterator.call(context, array[index], index, array) === Utils.BREAK) {
                    return Utils.BREAK;
                }
            }
        }
    };

    /**
     * Returns the array index of the first element in the passed array, that
     * passes a truth test.
     *
     * @param {Array} array
     *  The array to be searched for a matching element.
     *
     * @param {Function} iterator
     *  The iterator implementing the truth test. Receives the array element,
     *  the element index, and the complete array.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.sorted=false]
     *      If set to true, the elements in the passed array are assumed to be
     *      sorted in a way that no element that passes the truth test precedes
     *      an element that does not pass the truth test (all matching elements
     *      are following all non-matching elements). Internally, a binary
     *      search algorithm will be used to speed up the search process.
     *  @param {Object} [options.context]
     *      If specified, the iterator will be called with this context (the
     *      symbol 'this' will be bound to the context inside the iterator
     *      function).
     *
     * @returns {Number}
     *  The array index of the first matching element in the array; or -1 if no
     *  element was found.
     */
    Utils.findFirstIndex = function (array, iterator, options) {

        var context = Utils.getOption(options, 'context'),
            index = 0;

        if (Utils.getBooleanOption(options, 'sorted', false)) {
            // fast binary search
            var begin = 0, end = array.length;
            while (begin < end) {
                index = Math.floor((begin + end) / 2);
                if (iterator.call(context, array[index], index, array)) {
                    end = index;
                } else {
                    begin = index + 1;
                }
            }
            if (begin < array.length) { return begin; }
        } else {
            // linear search
            for (index = 0; index < array.length; index += 1) {
                if (iterator.call(context, array[index], index, array)) {
                    return index;
                }
            }
        }

        return -1;
    };

    /**
     * Returns the array index of the last element in the passed array, that
     * passes a truth test.
     *
     * @param {Array} array
     *  The array to be searched for a matching element.
     *
     * @param {Function} iterator
     *  The iterator implementing the truth test. Receives the array element,
     *  the element index, and the complete array. The method iterates
     *  backwards through the array, and returns the first array element that
     *  passes the truth test.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.sorted=false]
     *      If set to true, the elements in the passed array are assumed to be
     *      sorted in a way that no element that passes the truth test follows
     *      an element that does not pass the truth test (all non-matching
     *      elements are following all matching elements). Internally, a binary
     *      search algorithm will be used to speed up the search process.
     *  @param {Object} [options.context]
     *      If specified, the iterator will be called with this context (the
     *      symbol 'this' will be bound to the context inside the iterator
     *      function).
     *
     * @returns {Number}
     *  The array index of the last matching element in the array; or -1 if no
     *  element was found.
     */
    Utils.findLastIndex = function (array, iterator, options) {

        var context = Utils.getOption(options, 'context'),
            index = 0;

        if (Utils.getBooleanOption(options, 'sorted', false)) {
            // fast binary search
            var begin = array.length - 1, end = -1;
            while (begin > end) {
                index = Math.ceil((begin + end) / 2);
                if (iterator.call(context, array[index], index, array)) {
                    end = index;
                } else {
                    begin = index - 1;
                }
            }
            if (begin < array.length) { return begin; }
        } else {
            // linear search
            for (index = array.length - 1; index >= 0; index -= 1) {
                if (iterator.call(context, array[index], index, array)) {
                    return index;
                }
            }
        }

        return -1;
    };

    /**
     * Finds the first element in the passed array, that passed a truth test.
     *
     * @param {Array} array
     *  The array to be searched for a matching element.
     *
     * @param {Function} iterator
     *  The iterator implementing the truth test. Receives the array element,
     *  the element index, and the complete array.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.sorted=false]
     *      If set to true, the elements in the passed array are assumed to be
     *      sorted in a way that no element that passes the truth test precedes
     *      an element that does not pass the truth test (all matching elements
     *      are following all non-matching elements). Internally, a binary
     *      search algorithm will be used to speed up the search process.
     *  @param {Object} [options.context]
     *      If specified, the iterator will be called with this context (the
     *      symbol 'this' will be bound to the context inside the iterator
     *      function).
     *
     * @returns {Any}
     *  The last matching element in the array; or undefined, if no element
     *  passes the truth test.
     */
    Utils.findFirst = function (array, iterator, options) {
        var index = Utils.findFirstIndex(array, iterator, options);
        return (index >= 0) ? array[index] : undefined;
    };

    /**
     * Finds the last element in the passed array, that passed a truth test.
     *
     * @param {Array} array
     *  The array to be searched for a matching element.
     *
     * @param {Function} iterator
     *  The iterator implementing the truth test. Receives the array element,
     *  the element index, and the complete array. The method iterates
     *  backwards through the array, and returns the first array element that
     *  passes the truth test.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.sorted=false]
     *      If set to true, the elements in the passed array are assumed to be
     *      sorted in a way that no element that passes the truth test follows
     *      an element that does not pass the truth test (all non-matching
     *      elements are following all matching elements). Internally, a binary
     *      search algorithm will be used to speed up the search process.
     *  @param {Object} [options.context]
     *      If specified, the iterator will be called with this context (the
     *      symbol 'this' will be bound to the context inside the iterator
     *      function).
     *
     * @returns {Any}
     *  The last matching element in the array; or undefined, if no element
     *  passes the truth test.
     */
    Utils.findLast = function (array, iterator, options) {
        var index = Utils.findLastIndex(array, iterator, options);
        return (index >= 0) ? array[index] : undefined;
    };

    // calculation, conversion, string manipulation ---------------------------

    /**
     * Returns whether the Boolean representation of the passed values are
     * equal (whether both values are 'truthy', or both values are 'falsy').
     *
     * @param {Any} value1
     *  The first value.
     *
     * @param {Any} value2
     *  The second value.
     *
     * @returns {Boolean}
     *  Whether the Boolean representation of the passed values are equal.
     */
    Utils.boolEq = function (value1, value2) {
        value1 = !value1; // JSHint prevents to write this inline into the comparison
        value2 = !value2;
        return value1 === value2;
    };

    /**
     * Returns whether the Boolean representation of the passed values are
     * different (whether exactly one of the passed values is 'truthy', and the
     * other value is 'falsy').
     *
     * @param {Any} value1
     *  The first value.
     *
     * @param {Any} value2
     *  The second value.
     *
     * @returns {Boolean}
     *  Whether the Boolean representation of the passed values are different.
     */
    Utils.boolXor = function (value1, value2) {
        value1 = !value1; // JSHint prevents to write this inline into the comparison
        value2 = !value2;
        return value1 !== value2;
    };

    /**
     * Restricts the passed value to the specified numeric range.
     *
     * @param {Number} value
     *  The value to be restricted to the given range.
     *
     * @param {Number} min
     *  The lower border of the range.
     *
     * @param {Number} max
     *  The upper border of the range.
     *
     * @returns {Number}
     *  The passed value, if inside the given range, otherwise either the lower
     *  or upper border.
     */
    Utils.minMax = function (value, min, max) {
        return Math.min(Math.max(value, min), max);
    };

    /**
     * Rounds the passed floating-point number to the nearest multiple of the
     * specified precision.
     *
     * @param {Number} value
     *  The floating-point number to be rounded.
     *
     * @param {Number} precision
     *  The precision used to round the number. The value 1 will round to
     *  integers (exactly like the Math.round() method). Must be positive. If
     *  less than 1, must be the inverse of an integer to prevent further
     *  internal rounding errors.
     *
     * @returns {Number}
     *  The rounded number.
     */
    Utils.round = function (value, precision) {
        // Multiplication with small value may result in rounding errors (e.g.,
        // 227*0.1 results in 22.700000000000003), division by inverse value
        // works sometimes (e.g. 227/(1/0.1) results in 22.7), rounding the
        // inverse before division finally should work in all(?) cases, but
        // restricts valid precisions to inverses of integer numbers.
        value = Math.round((precision < 1) ? (value * Math.round(1 / precision)) : (value / precision));
        return (precision < 1) ? (value / Math.round(1 / precision)) : (value * precision);
    };

    /**
     * Rounds the passed floating-point number down to the nearest multiple of
     * the specified precision.
     *
     * @param {Number} value
     *  The floating-point number to be rounded.
     *
     * @param {Number} precision
     *  The precision used to round the number. The value 1 will round down to
     *  integers (exactly like the Math.floor() method). Must be positive. If
     *  less than 1, must be the inverse of an integer to prevent further
     *  internal rounding errors.
     *
     * @returns {Number}
     *  The rounded number.
     */
    Utils.roundDown = function (value, precision) {
        // see comment in Utils.round()
        value = Math.floor((precision < 1) ? (value * Math.round(1 / precision)) : (value / precision));
        return (precision < 1) ? (value / Math.round(1 / precision)) : (value * precision);
    };

    /**
     * Rounds the passed floating-point number up to the nearest multiple of
     * the specified precision.
     *
     * @param {Number} value
     *  The floating-point number to be rounded.
     *
     * @param {Number} precision
     *  The precision used to round the number. The value 1 will round up to
     *  integers (exactly like the Math.ceil() method). Must be positive. If
     *  less than 1, must be the inverse of an integer to prevent further
     *  internal rounding errors.
     *
     * @returns {Number}
     *  The rounded number.
     */
    Utils.roundUp = function (value, precision) {
        // see comment in Utils.round()
        value = Math.ceil((precision < 1) ? (value * Math.round(1 / precision)) : (value / precision));
        return (precision < 1) ? (value / Math.round(1 / precision)) : (value * precision);
    };

    /**
     * Rounds the passed floating-point number to the specified number of
     * significant digits, independent from the number of digits before and
     * after the decimal point.
     *
     * @param {Number} value
     *  The value to be rounded.
     *
     * @param {Number} digits
     *  The number of significant digits. Must be positive.
     *
     * @returns {Number}
     *  The rounded number.
     */
    Utils.roundSignificantDigits = function (value, digits) {
        var intDigits = Math.floor(Math.log(Math.abs(value)) / Math.LN10) + 1,
            exp = digits - intDigits,
            pow10 = (exp < 0) ? Math.pow(10, -exp) : Math.pow(10, exp);
        return (exp < 0) ? (Math.round(value / pow10) * pow10) : (Math.round(value * pow10) / pow10);
    };

    /**
     * Iterates through a range of numbers, without creating a temporary array.
     * Can be used as replacement for the for-loop to work-around the JSHint
     * warning that prevents to create local functions inside plain loops.
     *
     * @param {Number} begin
     *  The iteration will be started with this value.
     *
     * @param {Number} end
     *  The iteration will be stopped before this value will be reached
     *  (half-open range).
     *
     * @param {Function} iterator
     *  The iterator function that will be called for every value. Receives the
     *  current value as first parameter. If the iterator returns the
     *  Utils.BREAK object, the iteration process will be stopped immediately.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Object} [options.context]
     *      If specified, the iterator will be called with this context (the
     *      symbol 'this' will be bound to the context inside the iterator
     *      function).
     *  @param {Number} [options.step=1]
     *      If specified, the current value will be increased or decreased by
     *      this amount.
     *
     * @returns {Utils.BREAK|Undefined}
     *  A reference to the Utils.BREAK object, if the iterator has returned
     *  Utils.BREAK to stop the iteration process, otherwise undefined.
     */
    Utils.iterateRange = function (begin, end, iterator, options) {

        var // context for iterator function
            context = Utils.getOption(options, 'context'),
            // step value
            step = Utils.getNumberOption(options, 'step', 1),
            // the current value
            value = begin;

        while ((step > 0) ? (value < end) : (value > end)) {
            if (iterator.call(context, value) === Utils.BREAK) { return Utils.BREAK; }
            value += step;
        }
    };

    /**
     * Calculates the sum of the values returned by the iterator function
     * invoked for each element of the passed array or property of the passed
     * object. Can be used as convenience shortcut for the Underscore 'reduce'
     * method.
     *
     * @param {Array|Object} list
     *  The object or array whose properties/elements will be iterated.
     *
     * @param {Function} iterator
     *  The iterator function invoked for all elements/properties. Receives
     *  the current object property value or array element. Must return a
     *  number.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Object} [options.context]
     *      If specified, the iterator will be called with this context (the
     *      symbol 'this' will be bound to the context inside the iterator
     *      function).
     *
     * @returns {Number}
     *  The sum of all results returned by the iterator function.
     */
    Utils.getSum = function (list, iterator, options) {

        var context = Utils.getOption(options, 'context');

        return _.reduce(list, function (sum, entry) {
            return sum + iterator.call(context, entry);
        }, 0);
    };

    /**
     * Converts a length value from an absolute CSS measurement unit into
     * another absolute CSS measurement unit.
     *
     * @param {Number} value
     *  The length value to convert, as floating-point number.
     *
     * @param {String} fromUnit
     *  The CSS measurement unit of the passed value, as string. Supported
     *  units are 'px' (pixels), 'pc' (picas), 'pt' (points), 'in' (inches),
     *  'cm' (centimeters), and 'mm' (millimeters).
     *
     * @param {String} toUnit
     *  The target measurement unit.
     *
     * @param {Number} [precision]
     *  If specified, the resulting length will be rounded to the nearest
     *  multiple of this value. Must be positive.
     *
     * @returns {Number}
     *  The length value converted to the target measurement unit, as
     *  floating-point number.
     */
    Utils.convertLength = (function () {

        var // the conversion factors between pixels and other units
            FACTORS = {
                'px': 1,
                'pc': 1 / 9,
                'pt': 4 / 3,
                'in': 96,
                'cm': 96 / 2.54,
                'mm': 96 / 25.4
            };

        return function (value, fromUnit, toUnit, precision) {
            value *= (FACTORS[fromUnit] || 1) / (FACTORS[toUnit] || 1);
            return _.isFinite(precision) ? Utils.round(value, precision) : value;
        };
    }());

    /**
     * Converts a length value from an absolute CSS measurement unit into 1/100
     * of millimeters.
     *
     * @param {Number} value
     *  The length value to convert, as floating-point number.
     *
     * @param {String} fromUnit
     *  The CSS measurement unit of the passed value, as string. See method
     *  Utils.convertLength() for a list of supported units.
     *
     * @returns {Number}
     *  The length value converted to 1/100 of millimeters, as integer.
     */
    Utils.convertLengthToHmm = function (value, fromUnit) {
        return Math.round(Utils.convertLength(value, fromUnit, 'mm') * 100);
    };

    /**
     * Converts a length value from 1/100 of millimeters into an absolute CSS
     * measurement unit.
     *
     * @param {Number} value
     *  The length value in 1/100 of millimeters to convert, as integer.
     *
     * @param {String} toUnit
     *  The target measurement unit. See method Utils.convertLength() for a
     *  list of supported units.
     *
     * @param {Number} [precision]
     *  If specified, the resulting length will be rounded to the nearest
     *  multiple of this value. Must be positive.
     *
     * @returns {Number}
     *  The length value converted to the target measurement unit, as
     *  floating-point number.
     */
    Utils.convertHmmToLength = function (value, toUnit, precision) {
        return Utils.convertLength(value / 100, 'mm', toUnit, precision);
    };

    /**
     * Converts a CSS length value with measurement unit into a value of
     * another absolute CSS measurement unit.
     *
     * @param {String} valueAndUnit
     *  The value with its measurement unit to be converted, as string.
     *
     * @param {String} toUnit
     *  The target CSS measurement unit. See method Utils.convertLength() for
     *  a list of supported units.
     *
     * @param {Number} [precision]
     *  If specified, the resulting length will be rounded to the nearest
     *  multiple of this value. Must be positive.
     *
     * @returns {Number}
     *  The length value converted to the target measurement unit, as
     *  floating-point number.
     */
    Utils.convertCssLength = function (valueAndUnit, toUnit, precision) {
        var value = parseFloat(valueAndUnit);
        if (!_.isFinite(value)) {
            value = 0;
        }
        if (value && (valueAndUnit.length > 2)) {
            value = Utils.convertLength(value, valueAndUnit.substr(-2), toUnit, precision);
        }
        return value;
    };

    /**
     * Converts a CSS length value with measurement unit into 1/100 of
     * millimeters.
     *
     * @param {String} valueAndUnit
     *  The value with its measurement unit to be converted, as string.
     *
     * @returns {Number}
     *  The length value converted to 1/100 of millimeters, as integer.
     */
    Utils.convertCssLengthToHmm = function (valueAndUnit) {
        return Math.round(Utils.convertCssLength(valueAndUnit, 'mm') * 100);
    };

    /**
     * Converts a length value from 1/100 of millimeters into an absolute CSS
     * measurement unit and returns the CSS length with its unit as string.
     *
     * @param {Number} value
     *  The length value in 1/100 of millimeters to convert, as integer.
     *
     * @param {String} toUnit
     *  The target measurement unit. See method Utils.convertLength() for a
     *  list of supported units.
     *
     * @param {Number} [precision]
     *  If specified, the resulting length will be rounded to the nearest
     *  multiple of this value. Must be positive.
     *
     * @returns {String}
     *  The length value converted to the target measurement unit, followed by
     *  the unit name.
     */
    Utils.convertHmmToCssLength = function (value, toUnit, precision) {
        return Utils.convertHmmToLength(value, toUnit, precision) + toUnit;
    };

    /**
     * Returns whether the passed space-separated token list contains the
     * specified token.
     *
     * @param {String} list
     *  Space-separated list of tokens.
     *
     * @param {String} token
     *  The token to look up in the token list.
     *
     * @returns {Boolean}
     *  Whether the token is contained in the token list.
     */
    Utils.containsToken = function (list, token) {
        return _.contains(list.split(/\s+/), token);
    };

    /**
     * Inserts the specified token into a space-separated token list. The token
     * will not be inserted if it is already contained in the list.
     *
     * @param {String} list
     *  Space-separated list of tokens.
     *
     * @param {String} token
     *  The token to be inserted into the token list.
     *
     * @param {String} [nothing]
     *  If specified, the name of a token that represents a special 'nothing'
     *  or 'empty' state. If this token is contained in the passed token list,
     *  it will be removed.
     *
     * @returns {String}
     *  The new token list containing the specified token.
     */
    Utils.addToken = function (list, token, nothing) {
        var tokens = (list.length > 0) ? list.split(/\s+/) : [];
        if (_.isString(nothing)) {
            tokens = _.without(tokens, nothing);
        }
        if (!_.contains(tokens, token)) {
            tokens.push(token);
        }
        return tokens.join(' ');
    };

    /**
     * Removes the specified token from a space-separated token list.
     *
     * @param {String} list
     *  Space-separated list of tokens.
     *
     * @param {String} token
     *  The token to be removed from the token list.
     *
     * @param {String} [nothing]
     *  If specified, the name of a token that represents a special 'nothing'
     *  or 'empty' state. If the resulting token list is empty, this token will
     *  be inserted instead.
     *
     * @returns {String}
     *  The new token list without the specified token.
     */
    Utils.removeToken = function (list, token, nothing) {
        var tokens = (list.length > 0) ? _.without(list.split(/\s+/), token) : [];
        if (_.isString(nothing)) {
            if (tokens.length > 0) { tokens = _.without(tokens, nothing); }
            if (tokens.length === 0) { tokens.push(nothing); }
        }
        return tokens.join(' ');
    };

    /**
     * Inserts a token into or removes a token from the specified
     * space-separated token list.
     *
     * @param {String} list
     *  Space-separated list of tokens.
     *
     * @param {String} token
     *  The token to be inserted into or removed from the token list.
     *
     * @param {Boolean} state
     *  If set to true, the token will be inserted into the token list,
     *  otherwise removed from the token list.
     *
     * @param {String} [nothing]
     *  If specified, the name of a token that represents a special 'nothing'
     *  or 'empty' state.
     *
     * @returns {String}
     *  The new token list.
     */
    Utils.toggleToken = function (list, token, state, nothing) {
        return Utils[state ? 'addToken' : 'removeToken'](list, token, nothing);
    };

    /**
     * Returns the passed text where a substring has been replaced with another
     * string.
     *
     * @param {String} text
     *  The original text.
     *
     * @param {Number} start
     *  The start position of the substring to be replaced.
     *
     * @param {Number} end
     *  The end position of the substring to be replaced.
     *
     * @param {String} newText
     *  The new text to be inserted into the original text.
     *
     * @returns {String}
     *  The resulting text with the replaced substring.
     */
    Utils.replaceSubString = function (text, start, end, newText) {
        return text.substring(0, start) + newText + text.substring(end);
    };

    /**
     * Repeats the passed string.
     *
     * @param {String} text
     *  The text to be repeated.
     *
     * @param {Number} count
     *  The number of repetitions.
     *
     * @param {String} [sep='']
     *  A separator string that will be inserted between the repetitions of the
     *  passed string.
     *
     * @returns {String}
     *  The generated string.
     */
    Utils.repeatString = function (text, count, sep) {
        var result = '';
        sep = sep || '';
        text += sep;
        while (count > 0) {
            if (count & 1) { result += text; }
            text += text;
            count >>= 1;
        }
        // check existence; String.slice(0, -0) returns an empty string!
        return sep ? result.slice(0, -sep.length) : result;
    };

    /**
     * Trims non-printable characters and white-spaces at the beginning and end
     * of the passed string.
     *
     * @param {String} text
     *  The text to be trimmed.
     *
     * @returns {String}
     *  The trimmed text.
     */
    Utils.trimString = function (text) {
        return text.replace(/^[\x00-\x1f\x80-\x9f\s]+|[\x00-\x1f\x80-\x9f\s]+$/g, '');
    };

    /**
     * Replaces all non-printable characters (Unicode characters 0x00-0x1F and
     * 0x80-0x9F) in the passed string with space characters.
     *
     * @param {String} text
     *  The text to be cleaned from non-printable characters.
     *
     * @returns {String}
     *  The new text without non-printable characters.
     */
    Utils.cleanString = function (text) {
        return text.replace(/[\x00-\x1f\x80-\x9f]/g, ' ');
    };

    /**
     * Trims non-printable characters and white-spaces at the beginning and end
     * of the passed string, and replaces all embedded non-printable characters
     * (ASCII 0x00-0x1F) with space characters.
     *
     * @param {String} text
     *  The text to be trimmed and cleaned from non-printable characters.
     *
     * @returns {String}
     *  The new trimmed text without non-printable characters.
     */
    Utils.trimAndCleanString = function (text) {
        return Utils.cleanString(Utils.trimString(text));
    };

    /**
     * Returns the passed text with a capitalized first character.
     *
     * @param {String} text
     *  The text to be converted.
     *
     * @returns {String}
     *  The passed text with a capitalized first character.
     */
    Utils.capitalize = function (text) {
        return (text.length > 0) ? (text[0].toUpperCase() + text.slice(1)) : '';
    };

    /**
     * Returns the passed text with capitalized words.
     *
     * @param {String} text
     *  The text to be converted.
     *
     * @returns {String}
     *  The passed text with capitalized words.
     */
    Utils.capitalizeWords = function (text) {
        return _.map(text.split(' '), Utils.capitalize).join(' ');
    };

    /**
     * Escapes HTML mark-up characters (angle brackets, ampersand, double
     * quotes, and apostrophes) in the passed text.
     *
     * @param {String} text
     *  The text containing special HTML mark-up characters.
     *
     * @returns {String}
     *  The passed text with all mark-up characters escaped.
     */
    Utils.escapeHTML = function (text) {
        return Utils.cleanString(text)
            // replace the ampersand with the text &amp; (must be done first!)
            .replace(/&/g, '&amp;')
            // replace the left angle bracket with the text &lt;
            .replace(/</g, '&lt;')
            // replace the right angle bracket with the text &gt;
            .replace(/>/g, '&gt;')
            // replace the slash character with the text &#47;
            .replace(/\//g, '&#47;')
            // replace the double quote character with the text &quot;
            .replace(/"/g, '&quot;')
            // replace the apostrophe with the text &#39; (&apos; is not an HTML entity!)
            .replace(/'/g, '&#39;')
            // replace the non-breakable space with the text &nbsp;
            .replace(/\xa0/g, '&nbsp;');
    };

    /**
     * Cleans up script elements and event handlers to prevent Cross Site Scripting.
     * And removes position:fixed CSS style to avoid clipboard content being visible.
     *
     * @param {String} html
     *  The html String to clean up.
     *
     * @returns {String}
     *  The cleaned up html String.
     */
    Utils.cleanUpHtml = function (html) {
        var result;

        if (!html) { return null; }

        // remove event handler to prevent script injection and XSS
        result = html.replace(/<[^>]*\s(onload|onunload|onerror|onerrorupdate|onchange)=.*>/gi, '');

        // replace zero chars created by Firefox
        result = result.replace(/[\x00]/gi, '');

        // remove script elements
        result = $.parseHTML(result, null, false /*keepScripts*/);

        // remove position fixed style to avoid clipboard content being visible
        $(result).find('*').andSelf().css('position', '');

        return result;
    };

    /**
     * Converting the UTC date time string to a date string in local time (keeping
     * the final letter 'Z', that indicates UTC).
     *
     * @param {String} utcString
     *  The UTC date string in ISO 8601 format.
     *
     * @param {Object} [options]
     *  A map of options to control the properties of the returned string.
     *  @param {Boolean} [options.useMilliSeconds=false]
     *      Whether the milliseconds shall be part of the returned string.
     *      Default is, that the string is returned without milliseconds.
     *
     * @returns {String}
     *  The current local date in an ISO string representation (including ending
     *  'Z', which is not quite correct).
     */
    Utils.convertUTCDateToLocalDate = function (utcString, options) {

        var // whether the seconds can be set to '00' in the returned string
            useMilliSeconds = Utils.getBooleanOption(options, 'useMilliSeconds', false),
            // the local date object
            localDate = new Date(utcString),
            // the local date string
            localDateString = null,
            // the offset in hours between local time and UTC time
            offset = localDate.getTimezoneOffset() / 60,
            // the hours of the local date
            hours = localDate.getHours();

        // hours outside the range 0..23 will affect the date automatically
        localDate.setHours(hours - offset);

        localDateString = localDate.toISOString();

        if (!useMilliSeconds) {
            // removing all digits after the final '.' together with this '.'
            localDateString = localDateString.replace(/\.\d{3}Z$/, 'Z');
        }

        return localDateString;
    };

    /**
     * Returning the current UTC time ISO date format.
     * ISO date format: '2012-04-17T13:38:00Z'
     * 'Z' at the end for UTC (+00:00)
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.useSeconds=true]
     *      Whether the seconds can be set to '00' in the returned string.
     *      Default is, that the string is returned with seconds and milliseconds.
     *
     * @returns {String}
     *  The current UTC date in an ISO string representation.
     */
    Utils.getIsoDateString = function (options) {

        var // the current date in UTC time zone
            d = new Date().toISOString(),
            // whether the seconds can be set to '00' in the returned string
            useSeconds = Utils.getBooleanOption(options, 'useSeconds', true);

        if (!useSeconds) {
            // replacing all digits before the trailing Z and after the final ':' with 0
            d = d.replace(/:[\d\.]+?Z$/, ':00Z');
        }

        return d;
    };

    /**
     * Returning the current local time faked as UTC time in ISO date format.
     * This is necessary, because the round trip with MS Office requires this.
     * So local time '2012-04-17' and '13:38:00' results in a string in
     * ISO date format: '2012-04-17T13:38:00Z'
     * 'Z' at the end for UTC (+00:00)
     * So string contains always a shifted time, except in UTC time zone.
     *
     * This functions shifts the offset in the milliseconds number and then
     * converts to date object, so that there are no problems with offsets
     * in hours, where the day, month or even year might be adapted.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.useSeconds=true]
     *      Whether the seconds can be set to '00' in the returned string.
     *      Default is, that the string is returned with seconds and milliseconds.
     *
     * @returns {String}
     *  The current local time as UTC date in an ISO string representation.
     */
    Utils.getMSFormatIsoDateString = function (options) {

        var // the local date object
            localDate = new Date(),
            // the local date time in milliseconds
            localTime = localDate.getTime(),
            // the offset in milliseconds between local time and UTC time
            offset = localDate.getTimezoneOffset() * 60 * 1000,
            // the shifted local time in milliseconds
            shiftedTime = localTime - offset,
            // the date string in ISO format
            isoDateString = new Date(shiftedTime).toISOString();

        if (!Utils.getBooleanOption(options, 'useSeconds', true)) {
            // replacing all digits before the trailing Z and after the final ':' with 0
            isoDateString = isoDateString.replace(/:[\d\.]+?Z$/, ':00Z');
        }

        return isoDateString;
    };

    // options object ---------------------------------------------------------

    /**
     * Extracts a property value from the passed object. If the property does
     * not exist, returns the specified default value.
     *
     * @param {Object|Undefined} options
     *  An object containing some properties. May be undefined.
     *
     * @param {String} name
     *  The name of the property to be returned.
     *
     * @param [def]
     *  The default value returned when the options parameter is not an object,
     *  or if it does not contain the specified property.
     *
     * @returns
     *  The value of the specified property, or the default value.
     */
    Utils.getOption = function (options, name, def) {
        return (_.isObject(options) && (name in options)) ? options[name] : def;
    };

    /**
     * Extracts a string property from the passed object. If the property does
     * not exist, or is not a string, returns the specified default value.
     *
     * @param {Object|Undefined} options
     *  An object containing some properties. May be undefined.
     *
     * @param {String} name
     *  The name of the string property to be returned.
     *
     * @param [def]
     *  The default value returned when the options parameter is not an object,
     *  or if it does not contain the specified property, or if the property is
     *  not a string. May be any value (not only strings).
     *
     * @param {Boolean} [nonEmpty=false]
     *  If set to true, only non-empty strings will be returned from the
     *  options object. Empty strings will be replaced with the specified
     *  default value.
     *
     * @returns
     *  The value of the specified property, or the default value.
     */
    Utils.getStringOption = function (options, name, def, nonEmpty) {
        var value = Utils.getOption(options, name);
        return (_.isString(value) && (!nonEmpty || (value.length > 0))) ? value : def;
    };

    /**
     * Extracts a boolean property from the passed object. If the property does
     * not exist, or is not a boolean value, returns the specified default
     * value.
     *
     * @param {Object|Undefined} options
     *  An object containing some properties. May be undefined.
     *
     * @param {String} name
     *  The name of the boolean property to be returned.
     *
     * @param [def]
     *  The default value returned when the options parameter is not an object,
     *  or if it does not contain the specified property, or if the property is
     *  not a boolean value. May be any value (not only booleans).
     *
     * @returns
     *  The value of the specified property, or the default value.
     */
    Utils.getBooleanOption = function (options, name, def) {
        var value = Utils.getOption(options, name);
        return _.isBoolean(value) ? value : def;
    };

    /**
     * Extracts a floating-point property from the passed object. If the
     * property does not exist, or is not a number, returns the specified
     * default value.
     *
     * @param {Object|Undefined} options
     *  An object containing some properties. May be undefined.
     *
     * @param {String} name
     *  The name of the floating-point property to be returned.
     *
     * @param [def]
     *  The default value returned when the options parameter is not an object,
     *  or if it does not contain the specified property, or if the property is
     *  not a number. May be any value (not only numbers).
     *
     * @param [min]
     *  If specified and a number, set a lower bound for the returned value. Is
     *  not used, if neither the property nor the passed default are numbers.
     *
     * @param [max]
     *  If specified and a number, set an upper bound for the returned value.
     *  Is not used, if neither the property nor the passed default are
     *  numbers.
     *
     * @param [precision]
     *  If specified, the resulting number will be rounded to the nearest
     *  multiple of this value. Must be positive.
     *
     * @returns
     *  The value of the specified property, or the default value.
     */
    Utils.getNumberOption = function (options, name, def, min, max, precision) {
        var value = Utils.getOption(options, name);
        value = (_.isNumber(value) && _.isFinite(value)) ? value : def;
        if (_.isNumber(value) && _.isFinite(value)) {
            if (_.isNumber(min) && _.isFinite(min) && (value < min)) { value = min; }
            if (_.isNumber(max) && _.isFinite(max) && (value > max)) { value = max; }
            return (_.isNumber(precision) && _.isFinite(precision)) ? Utils.round(value, precision) : value;
        }
        return value;
    };

    /**
     * Extracts an integer property from the passed object. If the property
     * does not exist, or is not a number, returns the specified default value.
     *
     * @param {Object|Undefined} options
     *  An object containing some properties. May be undefined.
     *
     * @param {String} name
     *  The name of the integer property to be returned.
     *
     * @param [def]
     *  The default value returned when the options parameter is not an object,
     *  or if it does not contain the specified property, or if the property is
     *  not a number. May be any value (not only numbers).
     *
     * @param [min]
     *  If specified and a number, set a lower bound for the returned value. Is
     *  not used, if neither the property nor the passed default are numbers.
     *
     * @param [max]
     *  If specified and a number, set an upper bound for the returned value.
     *  Is not used, if neither the property nor the passed default are
     *  numbers.
     *
     * @returns
     *  The value of the specified property, or the default value, rounded to
     *  an integer.
     */
    Utils.getIntegerOption = function (options, name, def, min, max) {
        return Utils.getNumberOption(options, name, def, min, max, 1);
    };

    /**
     * Extracts an object property from the passed object. If the property does
     * not exist, or is not an object, returns the specified default value.
     *
     * @param {Object|Undefined} options
     *  An object containing some properties. May be undefined.
     *
     * @param {String} name
     *  The name of the property to be returned.
     *
     * @param [def]
     *  The default value returned when the options parameter is not an object,
     *  or if it does not contain the specified property, or if the property is
     *  not an object. May be any value (not only objects).
     *
     * @returns
     *  The value of the specified property, or the default value.
     */
    Utils.getObjectOption = function (options, name, def) {
        var value = Utils.getOption(options, name);
        return (_.isObject(value) && !_.isFunction(value) && !_.isArray(value) && !_.isRegExp(value)) ? value : def;
    };

    /**
     * Extracts a function from the passed object. If the property does not
     * exist, or is not a function, returns the specified default value.
     *
     * @param {Object|Undefined} options
     *  An object containing some properties. May be undefined.
     *
     * @param {String} name
     *  The name of the property to be returned.
     *
     * @param [def]
     *  The default value returned when the options parameter is not an object,
     *  or if it does not contain the specified property, or if the property is
     *  not an object. May be any value (not only functions).
     *
     * @returns
     *  The value of the specified property, or the default value.
     */
    Utils.getFunctionOption = function (options, name, def) {
        var value = Utils.getOption(options, name);
        return _.isFunction(value) ? value : def;
    };

    /**
     * Extracts a array from the passed object. If the property does not exist,
     * or is not an array, returns the specified default value.
     *
     * @param {Object|Undefined} options
     *  An object containing some properties. May be undefined.
     *
     * @param {String} name
     *  The name of the property to be returned.
     *
     * @param [def]
     *  The default value returned when the options parameter is not an object,
     *  or if it does not contain the specified property, or if the property is
     *  not an array. May be any value.
     *
     * @param {Boolean} [nonEmpty=false]
     *  If set to true, only non-empty arrays will be returned from the options
     *  object. Empty arrays will be replaced with the specified default value.
     *
     * @returns
     *  The value of the specified property, or the default value.
     */
    Utils.getArrayOption = function (options, name, def, nonEmpty) {
        var value = Utils.getOption(options, name);
        return (_.isArray(value) && (!nonEmpty || (value.length > 0))) ? value : def;
    };

    /**
     * Creates and returns a merged options map from the passed objects. Unlike
     * Underscore's extend() method, does not modify the passed objects, but
     * creates and returns a clone. Additionally, extends embedded plain JS
     * objects deeply instead of replacing them, for example, extending the
     * objects {a:{b:1}} and {a:{c:2}} will result in {a:{b:1,c:2}}.
     *
     * @param {Object} [...]
     *  One or more objects whose properties will be inserted into the
     *  resulting object.
     *
     * @returns {Object}
     *  A new object containing all properties of the passed objects.
     */
    Utils.extendOptions = function () {

        var // the resulting options
            result = {};

        function isPlainObject(value) {
            return _.isObject(value) && (value.constructor === Object);
        }

        function extend(options, extensions) {
            _.each(extensions, function (value, name) {
                if (isPlainObject(value)) {
                    // extension value is a plain object: ensure that the options map contains an embedded object
                    if (!isPlainObject(options[name])) {
                        options[name] = {};
                    }
                    extend(options[name], value);
                } else {
                    // extension value is not a plain object: clear old value, even if it was an object
                    options[name] = value;
                }
            });
        }

        // add all objects to the clone
        for (var index = 0; index < arguments.length; index += 1) {
            if (_.isObject(arguments[index])) {
                extend(result, arguments[index]);
            }
        }

        return result;
    };

    // generic DOM/CSS helpers ------------------------------------------------

    /**
     * A jQuery function selector that returns true if the DOM node bound to
     * the 'this' symbol is a text node. Can be used in all helper functions
     * that expect a jQuery selector including functions.
     */
    Utils.JQ_TEXTNODE_SELECTOR = function () { return this.nodeType === 3; };

    /**
     * Creates and returns a new <div> container element.
     *
     * @param {String} className
     *  CSS class name set at the container element.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {String} [options.classes]
     *      A space-separated list of CSS classes to be added to the element.
     *
     * @returns {jQuery}
     *  A jQuery object containing the new container element.
     */
    Utils.createContainerNode = function (className, options) {
        var classes = Utils.getStringOption(options, 'classes', '');
        return $('<div class="' + className + (classes ? ' ' : '') + classes + '">');
    };

    /**
     * Adds device-specific CSS marker classes to the specified element. The
     * following CSS classes will be added:
     * - 'touch', if the device has a touch surface.
     * - 'browser-firefox', if the current browser is Firefox.
     * - 'browser-webkit', if the current browser is based on WebKit.
     * - 'browser-chrome', if the current browser is Chrome.
     * - 'browser-safari', if the current browser is Safari.
     * - 'browser-ie', if the current browser is Internet Explorer.
     *
     * @param {HTMLElement|jQuery} node
     *  A single DOM element, or a jQuery collection with one or more elements.
     */
    Utils.addDeviceMarkers = function (node) {

        // convert to jQuery collection
        node = $(node);

        // marker for touch devices
        node.toggleClass('touch', Modernizr.touch);

        // marker for browser types
        node.toggleClass('browser-firefox', !!_.browser.Firefox)
            .toggleClass('browser-webkit', !!_.browser.WebKit)
            .toggleClass('browser-chrome', !!_.browser.Chrome)
            .toggleClass('browser-safari', !!_.browser.Safari)
            .toggleClass('browser-ie', !!_.browser.IE);
    };

    /**
     * Converts the passed object to a DOM node object.
     *
     * @param {Node|jQuery} node
     *  If the object is a DOM node object, returns it unmodified. If the
     *  object is a jQuery collection, returns its first node.
     *
     * @returns {Node}
     *  The DOM node object.
     */
    Utils.getDomNode = function (node) {
        return (node instanceof $) ? node.get(0) : node;
    };

    /**
     * Returns whether the passed node is a specific DOM element node.
     *
     * @param {Node|jQuery|Null|Undefined} node
     *  The DOM node to be checked. May be null or undefined.
     *
     * @param {String|Function|Node|jQuery} [selector]
     *  A jQuery selector that can be used to check the passed node for a
     *  specific type etc. The selector will be passed to the jQuery method
     *  jQuery.is(). If this selector is a function, it will be called with the
     *  DOM node bound to the symbol 'this'. See the jQuery API documentation
     *  at http://api.jquery.com/is for details.
     *
     * @returns {Boolean}
     *  Whether the passed node is an element node that matches the passed
     *  selector.
     */
    Utils.isElementNode = function (node, selector) {
        if (!node) { return false; }
        return (Utils.getDomNode(node).nodeType === 1) && (!selector || $(node).is(selector));
    };

    /**
     * Returns the lower-case name of a DOM node object.
     *
     * @param {Node|jQuery} node
     *  The DOM node whose name will be returned. If this object is a jQuery
     *  collection, uses the first node it contains.
     *
     * @returns {String}
     *  The lower-case name of the DOM node object. If the node is a text node,
     *  the string '#text' will be returned.
     */
    Utils.getNodeName = function (node) {
        return Utils.getDomNode(node).nodeName.toLowerCase();
    };

    /**
     * Returns whether the passed outer node contains the passed inner node.
     *
     * @param {Node|jQuery} outerNode
     *  The outer DOM node. If this object is a jQuery collection, uses the
     *  first node it contains.
     *
     * @param {Node|jQuery} innerNode
     *  The inner DOM node. If this object is a jQuery collection, uses the
     *  first node it contains.
     *
     * @returns {Boolean}
     *  Whether the inner node is a descendant of the outer node.
     */
    Utils.containsNode = function (outerNode, innerNode) {

        // convert nodes to DOM node object
        outerNode = Utils.getDomNode(outerNode);
        innerNode = Utils.getDomNode(innerNode);

        // IE does not support contains() method at document
        if (outerNode === document) {
            outerNode = document.body;
            if (innerNode === outerNode) {
                return true;
            }
        }

        // outer node must be an element; be sure that a node does not contain itself
        if ((outerNode.nodeType !== 1) || (outerNode === innerNode)) {
            return false;
        }

        // IE does not accept a DOM text node as parameter for the Node.contains() method
        if (innerNode.nodeType !== 1) {
            innerNode = innerNode.parentNode;
            // check if the parent of the inner text node is the outer node
            if (outerNode === innerNode) {
                return true;
            }
        }

        // Internet Explorer has no 'contains' at svg elements (26172)
        if (! _.isFunction(outerNode.contains)) {
            return $(innerNode).closest(outerNode).length > 0;
        }

        // use the native Node.contains() method
        return outerNode.contains(innerNode);
    };

    /**
     * Returns an integer attribute value of the passed element.
     *
     * @param {HTMLElement|jQuery} node
     *  The DOM element whose attribute will be returned. If this object is a
     *  jQuery collection, uses the first node it contains.
     *
     * @param {String} name
     *  The name of the element attribute.
     *
     * @param {Number} def
     *  A default value in case the element does not contain the specified
     *  attribute.
     *
     * @returns {Number}
     *  The attribute value parsed as integer, or the default value.
     */
    Utils.getElementAttributeAsInteger = function (node, name, def) {
        var attr = $(node).attr(name);
        return _.isString(attr) ? parseInt(attr, 10) : def;
    };

    /**
     * Returns whether the passed CSS border position is oriented vertically
     * (either 'top' or 'bottom').
     *
     * @param {String} position
     *  The CSS position, one of 'top', 'bottom', 'left', or 'right'.
     *
     * @returns {Boolean}
     *  Whether the passed position is either 'top' or 'bottom'.
     */
    Utils.isVerticalPosition = function (position) {
        return (position === 'top') || (position === 'bottom');
    };

    /**
     * Returns whether the passed CSS border position is the leading side
     * (either 'top' or 'left').
     *
     * @param {String} position
     *  The CSS position, one of 'top', 'bottom', 'left', or 'right'.
     *
     * @returns {Boolean}
     *  Whether the passed position is either 'top' or 'left'.
     */
    Utils.isLeadingPosition = function (position) {
        return (position === 'top') || (position === 'left');
    };

    /**
     * Sets a CSS formatting attribute with all browser-specific prefixes at
     * the passed element.
     *
     * @param {HTMLElement|jQuery} node
     *  The DOM element whose CSS attribute will be changed. If this object is
     *  a jQuery collection, changes all contained nodes.
     *
     * @param {String} name
     *  The base name of the CSS attribute.
     *
     * @param {Any} value
     *  The new value of the CSS attribute.
     */
    Utils.setCssAttributeWithPrefixes = (function () {

        var // the prefix for the current browser
            prefix = _.browser.WebKit ? '-webkit-' : _.browser.Firefox ? '-moz-' : _.browser.IE ? '-ms-' : '';

        return function (node, name, value) {
            var props = Utils.makeSimpleObject(name, value);
            if (prefix) { props[prefix + name] = value; }
            $(node).css(props);
        };
    }());

    /**
     * Returns an integer indicating how the two passed nodes are located to
     * each other.
     *
     * @param {Node|jQuery} node1
     *  The first DOM node tested if it is located before the second node. If
     *  this object is a jQuery collection, uses the first node it contains.
     *
     * @param {Node|jQuery} node2
     *  The second DOM node. If this object is a jQuery collection, uses the
     *  first node it contains.
     *
     * @returns {Number}
     *  The value zero, if the nodes are equal, a negative number, if node1
     *  precedes or contains node2, or a positive number, if node2 precedes or
     *  contains node1.
     */
    Utils.compareNodes = function (node1, node2) {
        node1 = Utils.getDomNode(node1);
        node2 = Utils.getDomNode(node2);
        return (node1 === node2) ? 0 : ((node1.compareDocumentPosition(node2) & 4) === 4) ? -1 : 1;
    };

    /**
     * Iterates over all descendant DOM nodes of the specified element.
     *
     * @param {HTMLElement|jQuery} element
     *  A DOM element object whose descendant nodes will be iterated. If this
     *  object is a jQuery collection, uses the first node it contains.
     *
     * @param {Function} iterator
     *  The iterator function that will be called for every node. Receives the
     *  DOM node as first parameter. If the iterator returns the Utils.BREAK
     *  object, the iteration process will be stopped immediately. The iterator
     *  can remove the visited node descendants from the DOM. It is also
     *  allowed to remove any siblings of the visited node, as long as the
     *  visited node will not be removed itself.
     *
     * @param {Object} [context]
     *  If specified, the iterator will be called with this context (the symbol
     *  'this' will be bound to the context inside the iterator function).
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.children=false]
     *      If set to true, only direct child nodes will be visited.
     *  @param {Boolean} [options.reverse=false]
     *      If set to true, the descendant nodes are visited in reversed order.
     *
     * @returns {Utils.BREAK|Undefined}
     *  A reference to the Utils.BREAK object, if the iterator has returned
     *  Utils.BREAK to stop the iteration process, otherwise undefined.
     */
    Utils.iterateDescendantNodes = function (element, iterator, context, options) {

        var // only child nodes
            childrenOnly = Utils.getBooleanOption(options, 'children', false),
            // iteration direction
            reverse = Utils.getBooleanOption(options, 'reverse', false),
            // visited node, and the next or previous sibling of the visited node
            childNode = null, nextSibling = null;

        // visit all child nodes
        element = Utils.getDomNode(element);
        for (childNode = reverse ? element.lastChild : element.firstChild; childNode; childNode = nextSibling) {

            // get next/previous sibling in case the iterator removes the node
            nextSibling = reverse ? childNode.previousSibling : childNode.nextSibling;

            // call iterator for child node; if it returns Utils.BREAK, exit loop and return too
            if (iterator.call(context, childNode) === Utils.BREAK) { return Utils.BREAK; }

            // iterate grand child nodes (only if the iterator did not remove the node from the DOM)
            if (Utils.containsNode(element, childNode)) {

                // refresh next sibling (iterator may have removed the old one, or may have inserted another one)
                nextSibling = reverse ? childNode.previousSibling : childNode.nextSibling;

                // if iterator for any descendant node returns Utils.BREAK, return too
                if (!childrenOnly && (childNode.nodeType === 1) && (Utils.iterateDescendantNodes(childNode, iterator, context, options) === Utils.BREAK)) {
                    return Utils.BREAK;
                }
            }
        }
    };

    /**
     * Iterates over selected descendant DOM nodes of the specified element.
     *
     * @param {HTMLElement|jQuery} element
     *  A DOM element object whose descendant nodes will be iterated. If this
     *  object is a jQuery collection, uses the first node it contains.
     *
     * @param {String|Function|Node|jQuery} selector
     *  A jQuery selector that will be used to decide whether to call the
     *  iterator function for the current DOM node. The selector will be passed
     *  to the jQuery method jQuery.is() for each node. If this selector is a
     *  function, it will be called with the current DOM node bound to the
     *  symbol 'this'. See the jQuery API documentation at
     *  http://api.jquery.com/is for details.
     *
     * @param {Function} iterator
     *  The iterator function that will be called for every matching node.
     *  Receives the DOM node as first parameter. If the iterator returns the
     *  Utils.BREAK object, the iteration process will be stopped immediately.
     *  The iterator can remove visited nodes from the DOM.
     *
     * @param {Object} [context]
     *  If specified, the iterator will be called with this context (the symbol
     *  'this' will be bound to the context inside the iterator function).
     *
     * @param {Object} [options]
     *  Optional parameters. Supports all options that are supported by the
     *  method Utils.iterateDescendantNodes().
     *
     * @returns {Utils.BREAK|Undefined}
     *  A reference to the Utils.BREAK object, if the iterator has returned
     *  Utils.BREAK to stop the iteration process, otherwise undefined.
     */
    Utils.iterateSelectedDescendantNodes = function (element, selector, iterator, context, options) {

        return Utils.iterateDescendantNodes(element, function (node) {
            if ($(node).is(selector) && (iterator.call(context, node) === Utils.BREAK)) {
                return Utils.BREAK;
            }
        }, context, options);
    };

    /**
     * Returns the first descendant DOM node in the specified element that
     * passes a truth test using the specified jQuery selector.
     *
     * @param {HTMLElement|jQuery} element
     *  A DOM element object whose descendant nodes will be searched. If this
     *  object is a jQuery collection, uses the first node it contains.
     *
     * @param {String|Function|Node|jQuery} selector
     *  A jQuery selector that will be used to find the DOM node. The selector
     *  will be passed to the jQuery method jQuery.is() for each node. If this
     *  selector is a function, it will be called with the current DOM node
     *  bound to the symbol 'this'. See the jQuery API documentation at
     *  http://api.jquery.com/is for details.
     *
     * @param {Object} [options]
     *  Optional parameters. Supports all options that are supported by the
     *  method Utils.iterateDescendantNodes().
     *
     * @returns {Node|Null}
     *  The first descendant DOM node that matches the passed selector. If no
     *  matching node has been found, returns null.
     */
    Utils.findDescendantNode = function (element, selector, options) {

        var // the node to be returned
            resultNode = null;

        // find the first node passing the iterator test
        Utils.iterateSelectedDescendantNodes(element, selector, function (node) {
            resultNode = node;
            return Utils.BREAK;
        }, undefined, options);

        return resultNode;
    };

    /**
     * Returns a child node of the passed node, that is at a specific index in
     * the array of all matching child nodes.
     *
     * @param {HTMLElement|jQuery} element
     *  A DOM element object whose child nodes will be visited. If this object
     *  is a jQuery collection, uses the first node it contains.
     *
     * @param {String|Function|Node|jQuery} selector
     *  A jQuery selector that will be used to decide which child nodes are
     *  matching while searching to the specified index. The selector will be
     *  passed to the jQuery method jQuery.is() for each node. If this selector
     *  is a function, it will be called with the current DOM node bound to the
     *  symbol 'this'. See the jQuery API documentation at
     *  http://api.jquery.com/is for details.
     *
     * @param {Number} index
     *  The zero-based index of the child node in the set of child nodes
     *  matching the selector that will be returned.
     *
     * @returns {Node|Null}
     *  The 'index'-th child node that matches the selector; or null, if the
     *  index is outside the valid range.
     */
    Utils.getSelectedChildNodeByIndex = function (element, selector, index) {

        var // the node to be returned
            resultNode = null;

        // find the 'index'-th matching child node
        Utils.iterateSelectedDescendantNodes(element, selector, function (node) {
            // node found: store and escape from loop
            if (index === 0) {
                resultNode = node;
                return Utils.BREAK;
            }
            index -= 1;
        }, undefined, { children: true });

        return resultNode;
    };

    /**
     * Finds the closest ancestor of the passed node that matches the specified
     * jQuery selector. If the specified node itself matches the selector, it
     * will be returned. In difference to the jQuery method jQuery.closest(),
     * the passed node selector can be a function.
     *
     * @param {HTMLElement|jQuery} rootNode
     *  The DOM root node that will not be left while searching for a matching
     *  node. If this object is a jQuery collection, uses the first node it
     *  contains.
     *
     * @param {Node|jQuery} node
     *  The DOM node whose chain of ancestors will be searched for a matching
     *  node. Must be a descendant of the passed root node. If this object is a
     *  jQuery collection, uses the first node it contains.
     *
     * @param {String|Function|Node|jQuery} selector
     *  A jQuery selector that will be used to find a matching DOM node. The
     *  selector will be passed to the jQuery method jQuery.is() for each node.
     *  If this selector is a function, it will be called with the current DOM
     *  node bound to the symbol 'this'. See the jQuery API documentation at
     *  http://api.jquery.com/is for details.
     *
     * @returns {Node|Null}
     *  The passed node, or the first ancestor node that matches the passed
     *  selector, and is contained in or equal to the root node; or null, no
     *  node has been found.
     */
    Utils.findClosest = function (rootNode, node, selector) {

        rootNode = Utils.getDomNode(rootNode);
        node = Utils.getDomNode(node);

        do {
            if ($(node).is(selector)) { return node; }
            if (node === rootNode) { return null; }
            node = node.parentNode;
        } while (node);

        return null;
    };

    /**
     * Finds the farthest ancestor of the passed node that matches the
     * specified jQuery selector. If the specified node itself matches the
     * selector but none of its ancestors, the node will be returned itself. In
     * difference to most jQuery methods, the passed node selector can be a
     * function.
     *
     * @param {HTMLElement|jQuery} rootNode
     *  The DOM root node that will not be left while searching for matching
     *  nodes. If this object is a jQuery collection, uses the first node it
     *  contains.
     *
     * @param {Node|jQuery} node
     *  The DOM node whose chain of ancestors will be searched for a matching
     *  node. Must be a descendant of the passed root node. If this object is a
     *  jQuery collection, uses the first node it contains.
     *
     * @param {String|Function|Node|jQuery} selector
     *  A jQuery selector that will be used to find a matching DOM node. The
     *  selector will be passed to the jQuery method jQuery.is() for each node.
     *  If this selector is a function, it will be called with the current DOM
     *  node bound to the symbol 'this'. See the jQuery API documentation at
     *  http://api.jquery.com/is for details.
     *
     * @returns {Node|Null}
     *  The passed node, or the farthest ancestor node that matches the passed
     *  selector, and is contained in or equal to the root node; or null, no
     *  node has been found.
     */
    Utils.findFarthest = function (rootNode, node, selector) {

        var // the last found matching node
            matchingNode = null;

        rootNode = Utils.getDomNode(rootNode);
        node = Utils.getDomNode(node);

        do {
            if ($(node).is(selector)) { matchingNode = node; }
            if (node === rootNode) { return matchingNode; }
            node = node.parentNode;
        } while (node);

        return null;
    };

    /**
     * Finds the closest previous sibling node of the passed node that matches
     * a jQuery selector.
     *
     * @param {Node|jQuery} node
     *  The DOM node whose previous matching sibling will be returned. If this
     *  object is a jQuery collection, uses the first node it contains.
     *
     * @param {String|Function|Node|jQuery} [selector]
     *  A jQuery selector that will be used to find a matching sibling. The
     *  selector will be passed to the jQuery method jQuery.is() for each node.
     *  If this selector is a function, it will be called with the current DOM
     *  node bound to the symbol 'this'. See the jQuery API documentation at
     *  http://api.jquery.com/is for details. If omitted, this method returns
     *  the direct previous sibling of the passed node.
     *
     * @returns {Node|Null}
     *  The previous matching sibling of the passed node; or null, no previous
     *  sibling node has been found.
     */
    Utils.findPreviousSiblingNode = function (node, selector) {
        node = Utils.getDomNode(node).previousSibling;
        if (!_.isUndefined(selector)) {
            while (node && !$(node).is(selector)) {
                node = node.previousSibling;
            }
        }
        return node;
    };

    /**
     * Finds the closest next sibling node of the passed node that matches a
     * jQuery selector.
     *
     * @param {Node|jQuery} node
     *  The DOM node whose next matching sibling will be returned. If this
     *  object is a jQuery collection, uses the first node it contains.
     *
     * @param {String|Function|Node|jQuery} [selector]
     *  A jQuery selector that will be used to find a matching sibling. The
     *  selector will be passed to the jQuery method jQuery.is() for each node.
     *  If this selector is a function, it will be called with the current DOM
     *  node bound to the symbol 'this'. See the jQuery API documentation at
     *  http://api.jquery.com/is for details. If omitted, this method returns
     *  the direct next sibling of the passed node.
     *
     * @returns {Node|Null}
     *  The next matching sibling of the passed node; or null, no next sibling
     *  node has been found.
     */
    Utils.findNextSiblingNode = function (node, selector) {
        node = Utils.getDomNode(node).nextSibling;
        if (!_.isUndefined(selector)) {
            while (node && !$(node).is(selector)) {
                node = node.nextSibling;
            }
        }
        return node;
    };

    /**
     * Finds a previous sibling node of the passed node in the DOM tree. If the
     * node is the first child of its parent, bubbles up the chain of parent
     * nodes and continues with the previous sibling of the first ancestor that
     * has more siblings. An optional jQuery selector can be specified to
     * search through the previous siblings until a matching node has been
     * found.
     *
     * @param {HTMLElement|jQuery} rootNode
     *  The DOM root node whose sub node tree will not be left when searching
     *  for the previous sibling node. If this object is a jQuery collection,
     *  uses the first node it contains.
     *
     * @param {Node|jQuery} node
     *  The DOM node whose previous sibling will be returned. Must be a
     *  descendant of the passed root node. If this object is a jQuery
     *  collection, uses the first node it contains.
     *
     * @param {String|Function|Node|jQuery} [selector]
     *  A jQuery selector that will be used to find a DOM node. The selector
     *  will be passed to the jQuery method jQuery.is() for each node. If this
     *  selector is a function, it will be called with the current DOM node
     *  bound to the symbol 'this'. See the jQuery API documentation at
     *  http://api.jquery.com/is for details. If omitted, this method returns
     *  the direct next sibling of the passed node.
     *
     * @param {String|Function|Node|jQuery} [skipSiblingsSelector]
     *  A jQuery selector that will be used to to identify DOM nodes whose
     *  siblings will be ignored completely, even if they match the specified
     *  selector. If omitted, this method visits all descendant nodes of all
     *  siblings while searching for a matching node.
     *
     * @returns {Node|Null}
     *  The previous sibling of the passed node in the DOM sub tree of the root
     *  node; or null, no node has been found.
     */
    Utils.findPreviousNode = function (rootNode, node, selector, skipSiblingsSelector) {

        // set 'node' to the previous sibling (of itself or one of its parents)
        function previousNode() {

            // go to previous sibling if existing
            if (node && node.previousSibling) {
                node = node.previousSibling;

                // start with deepest last descendant node in the previous sibling
                while (node && (_.isUndefined(skipSiblingsSelector) || !$(node).is(skipSiblingsSelector)) && node.lastChild) {
                    node = node.lastChild;
                }
            } else {
                // otherwise, go to parent, but do not leave the root node
                node = (node.parentNode === rootNode) ? null : node.parentNode;
            }
        }

        rootNode = Utils.getDomNode(rootNode);
        node = Utils.getDomNode(node);

        previousNode();
        if (!_.isUndefined(selector)) {
            while (node && !$(node).is(selector)) { previousNode(); }
        }

        return node;
    };

    /**
     * Finds a next sibling node of the passed node in the DOM tree. If the
     * node is the last child of its parent, bubbles up the chain of parent
     * nodes and continues with the next sibling of the first ancestor that has
     * more siblings. An optional jQuery selector can be specified to search
     * through the next siblings until a matching node has been found.
     *
     * @param {HTMLElement|jQuery} rootNode
     *  The DOM root node whose sub node tree will not be left when searching
     *  for the next sibling node. If this object is a jQuery collection, uses
     *  the first node it contains.
     *
     * @param {Node|jQuery} node
     *  The DOM node whose next sibling will be returned. Must be a descendant
     *  of the passed root node. If this object is a jQuery collection, uses
     *  the first node it contains.
     *
     * @param {String|Function|Node|jQuery} [selector]
     *  A jQuery selector that will be used to find a DOM node. The selector
     *  will be passed to the jQuery method jQuery.is() for each node. If this
     *  selector is a function, it will be called with the current DOM node
     *  bound to the symbol 'this'. See the jQuery API documentation at
     *  http://api.jquery.com/is for details. If omitted, this method returns
     *  the direct next sibling of the passed node.
     *
     * @param {String|Function|Node|jQuery} [skipSiblingsSelector]
     *  A jQuery selector that will be used to to identify DOM nodes whose
     *  siblings will be ignored completely, even if they match the specified
     *  selector. If omitted, this method visits all descendant nodes of all
     *  siblings while searching for a matching node.
     *
     * @returns {Node|Null}
     *  The next sibling of the passed node in the DOM sub tree of the root
     *  node; or null, no node has been found.
     */
    Utils.findNextNode = function (rootNode, node, selector, skipSiblingsSelector) {

        var // visit descendant nodes
            visitDescendants = false;

        // set 'node' to the next sibling (of itself or one of its parents)
        function nextNode() {

            // visit descendant nodes if specified
            if (visitDescendants && node && (_.isUndefined(skipSiblingsSelector) || !$(node).is(skipSiblingsSelector)) && node.firstChild) {
                node = node.firstChild;
                return;
            }

            // for next call: always visit descendant nodes of own or parent siblings
            visitDescendants = true;

            // find first node up the tree that has a next sibling
            while (node && !node.nextSibling) {
                // do not leave the root node
                node = (node.parentNode === rootNode) ? null : node.parentNode;
            }

            // go to that next sibling
            node = node && node.nextSibling;
        }

        rootNode = Utils.getDomNode(rootNode);
        node = Utils.getDomNode(node);

        nextNode();
        if (!_.isUndefined(selector)) {
            while (node && !$(node).is(selector)) { nextNode(); }
        }

        return node;
    };

    // positioning and scrolling ----------------------------------------------

    // calculate size of system scroll bars
    (function () {

        var // dummy container used to calculate the scroll bar sizes
            node = $('<div>').css({ width: '100px', height: '100px', overflow: 'scroll' });

        $('body').append(node);
        Utils.SCROLLBAR_WIDTH = node.width() - node[0].clientWidth;
        Utils.SCROLLBAR_HEIGHT = node.height() - node[0].clientHeight;
        node.remove();
    }());

    /**
     * Returns whether the passed DOM element contains a visible vertical
     * scroll bar.
     *
     * @param {HTMLElement|jQuery} node
     *  The DOM element. If this object is a jQuery collection, uses the first
     *  node it contains.
     *
     * @returns {Boolean}
     *  Whether the DOM element contains a visible vertical scroll bar.
     */
    Utils.hasVerticalScrollBar = function (node) {
        return $(node).width() > Utils.getDomNode(node).clientWidth;
    };

    /**
     * Returns whether the passed DOM element contains a visible horizontal
     * scroll bar.
     *
     * @param {HTMLElement|jQuery} node
     *  The DOM element. If this object is a jQuery collection, uses the first
     *  node it contains.
     *
     * @returns {Boolean}
     *  Whether the DOM element contains a visible horizontal scroll bar.
     */
    Utils.hasHorizontalScrollBar = function (node) {
        return $(node).height() > Utils.getDomNode(node).clientHeight;
    };

    /**
     * Returns the size of the entire document page.
     *
     * @returns {Object}
     *  An object with numeric properties representing the size of the document
     *  page, in pixels:
     *  - {Number} width
     *      The width of the document page.
     *  - {Number} height
     *      The height of the document page.
     */
    Utils.getPageSize = function () {

        var // the OX root node
            coreNode = $('#io-ox-core')[0];

        return { width: coreNode.offsetWidth, height: coreNode.offsetHeight };
    };

    /**
     * Returns the exact size of the passed node, as floating-point numbers.
     *
     * @param {HTMLElement|jQuery} node
     *  The DOM element whose size will be returned. If this object is a jQuery
     *  collection, uses the first node it contains.
     *
     * @returns {Object}
     *  An object with numeric properties representing the size of the node, in
     *  floating-point pixels:
     *  - {Number} width
     *      The outer width of the node (including its borders).
     *  - {Number} height
     *      The outer height of the node (including its borders).
     */
    Utils.getExactNodeSize = function (node) {
        var boundingRect = Utils.getDomNode(node).getBoundingClientRect();
        return { width: boundingRect.width, height: boundingRect.height };
    };

    /**
     * Returns the position and size of the specified node inside the entire
     * document page. This includes the distances of all four borders of the
     * node to the borders of the document.
     *
     * @param {HTMLElement|jQuery} node
     *  The DOM element whose position relative to the document page will be
     *  calculated. If this object is a jQuery collection, uses the first node
     *  it contains.
     *
     * @returns {Object}
     *  An object with numeric properties representing the position and size of
     *  the node relative to the document page, in pixels:
     *  - {Number} left
     *      The distance of the left border of the node to the left border of
     *      the document page.
     *  - {Number} top
     *      The distance of the top border of the node to the top border of the
     *      document page.
     *  - {Number} right
     *      The distance of the right border of the node to the right border of
     *      the document page.
     *  - {Number} bottom
     *      The distance of the bottom border of the node to the bottom border
     *      of the document page.
     *  - {Number} width
     *      The outer width of the node (including its borders).
     *  - {Number} height
     *      The outer height of the node (including its borders).
     */
    Utils.getNodePositionInPage = function (node) {

        var // the passed node, as jQuery object
            $node = $(node),
            // the offset of the node, relative to the browser window
            position = $node.offset(),
            // the size of the entire document page
            pageSize = Utils.getPageSize();

        position.width = $node.outerWidth();
        position.height = $node.outerHeight();
        position.right = pageSize.width - position.left - position.width;
        position.bottom = pageSize.height - position.top - position.height;

        return position;
    };

    /**
     * Returns the position of the visible area of the passed container node.
     * This includes the size of the visible area without scroll bars (if
     * shown), and the distances of all four borders of the visible area to the
     * borders of the entire scroll area.
     *
     * @param {HTMLElement|jQuery} containerNode
     *  The DOM container element. If this object is a jQuery collection, uses
     *  the first node it contains.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.regardSoftkeyboard=false]
     *      special flag for upcoming softkeyboard, if set to true,
     *      the containerNodes clientheight will be interpreted smaller,
     *      that we better scroll to the visible area
     *
     * @returns {Object}
     *  An object with numeric properties representing the position and size of
     *  the visible area relative to the entire scroll area of the container
     *  node in pixels:
     *  - {Number} left
     *      The distance of the left border of the visible area to the left
     *      border of the entire scroll area.
     *  - {Number} top
     *      The distance of the top border of the visible area to the top
     *      border of the entire scroll area.
     *  - {Number} right
     *      The distance of the right border of the visible area (without
     *      scroll bar) to the right border of the entire scroll area.
     *  - {Number} bottom
     *      The distance of the bottom border of the visible area (without
     *      scroll bar) to the bottom border of the entire scroll area.
     *  - {Number} width
     *      The width of the visible area (without scroll bar).
     *  - {Number} height
     *      The height of the visible area (without scroll bar).
     */
    Utils.getVisibleAreaPosition = function (containerNode, options) {
        containerNode = Utils.getDomNode(containerNode);

        var softKeyboardHeight = 0;

        if (Utils.getBooleanOption(options, 'regardSoftkeyboard', false)) {
            //here comes the special code for softkeyboards
            //if the touch lands in the lower area we fake that its outsice of the client-rect,
            //so we scroll before the browser scroll the complete body



            //special for ipad, because it wants to scroll always unasked...
            if (Utils.IPAD) { window.scrollTo(0, 0); }

            var relKeyHeight = 0;
            var screenHeight;
            var portrait;

            if ( _.browser.iOS) {
                portrait = Math.abs(window.orientation) !== 90;
            } else {
                portrait = screen.height > screen.width;
            }

            if (portrait) {
                if (Utils.IPAD){ relKeyHeight = 2/3; }
                else { relKeyHeight = 3/4; }
                screenHeight = Math.max(screen.height, screen.width);
            } else {
                relKeyHeight = 1/2;
                screenHeight = Math.min(screen.height, screen.width);
            }

            if (window.innerHeight >= (screenHeight * relKeyHeight)) {
                softKeyboardHeight = Math.round((1-relKeyHeight) * screen.height);
            }
        }

        var clientHeight = containerNode.clientHeight - softKeyboardHeight;

        return {
            left: containerNode.scrollLeft,
            top: containerNode.scrollTop,
            right: containerNode.scrollWidth - containerNode.clientWidth - containerNode.scrollLeft,
            bottom: containerNode.scrollHeight - clientHeight - containerNode.scrollTop,
            width: containerNode.clientWidth,
            height: clientHeight
        };
    };

    /**
     * Returns the position and size of the specified document page rectangle
     * inside the passed DOM container node. This includes the size of the
     * rectangle as specified, and the distances of all four borders of the
     * rectangle to the borders of the visible area or entire scroll area of
     * the container node.
     *
     * @param {HTMLElement|jQuery} containerNode
     *  The container DOM element. If this object is a jQuery collection, uses
     *  the first node it contains.
     *
     * @param {Object} pageRect
     *  The document page rectangle whose position will be converted relatively
     *  to the position of the container node. Must provide the properties
     *  'left', 'top', 'width', and 'height' in pixels. The properties 'left'
     *  and 'top' are interpreted relatively to the entire document page.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.visibleArea=false]
     *      If set to true, calculates the distances of the rectangle to the
     *      visible area of the container node. Otherwise, calculates the
     *      distances of the rectangle to the entire scroll area of the
     *      container node.
     *  @param {Boolean} [options.regardSoftkeyboard=false]
     *      special flag for upcoming softkeyboard, if set to true,
     *      the containerNodes clientheight will be interpreted smaller,
     *      that we better scroll to the visible area
     *
     * @returns {Object}
     *  An object with numeric properties representing the position and size of
     *  the page rectangle relative to the entire scroll area or visible area
     *  of the container node, in pixels:
     *  - {Number} left
     *      The distance of the left border of the rectangle to the left border
     *      of the container node.
     *  - {Number} top
     *      The distance of the top border of the rectangle to the top border
     *      of the container node.
     *  - {Number} right
     *      The distance of the right border of the rectangle to the right
     *      border of the container node.
     *  - {Number} bottom
     *      The distance of the bottom border of the rectangle to the bottom
     *      border of the container node.
     *  - {Number} width
     *      The width of the rectangle, as passed.
     *  - {Number} height
     *      The height of the rectangle, as passed.
     */
    Utils.getRectanglePositionInNode = function (containerNode, pageRect, options) {

        var // the passed container node, as jQuery object
            $containerNode = $(containerNode),
            // the offset of the container node, relative to the document body
            containerOffset = containerNode.offset(),
            // the width of the left and top border of the scrollable node, in pixels
            leftBorderWidth = Utils.convertCssLength($containerNode.css('borderLeftWidth'), 'px'),
            topBorderWidth = Utils.convertCssLength($containerNode.css('borderTopWidth'), 'px'),
            // position and size of the visible area of the container node
            visiblePosition = Utils.getVisibleAreaPosition(containerNode, options),

            // the position and size, relative to the visible area of the container node
            position = {
                left: pageRect.left + leftBorderWidth - containerOffset.left,
                top: pageRect.top + topBorderWidth - containerOffset.top,
                width: pageRect.width,
                height: pageRect.height
            };

        // add right and bottom distance of child node to visible area
        position.right = visiblePosition.width - position.left - position.width;
        position.bottom = visiblePosition.height - position.top - position.height;

        // add distances to entire scroll area, if option 'visibleArea' is not set
        if (!Utils.getBooleanOption(options, 'visibleArea', false)) {
            _.each(['left', 'top', 'right', 'bottom'], function (border) {
                position[border] += visiblePosition[border];
            });
        }

        return position;
    };

    /**
     * Returns the position and size of the specified child node inside its
     * ancestor container node. This includes the outer size of the child node,
     * and the distances of all four borders of the child node to the borders
     * of the visible area or entire scroll area of the container node.
     *
     * @param {HTMLElement|jQuery} containerNode
     *  The DOM container element. If this object is a jQuery collection, uses
     *  the first node it contains.
     *
     * @param {HTMLElement|jQuery} childNode
     *  The DOM element whose dimensions will be calculated. Must be contained
     *  in the specified container element. If this object is a jQuery
     *  collection, uses the first node it contains.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.visibleArea=false]
     *      If set to true, calculates the distances of the child node to the
     *      visible area of the container node. Otherwise, calculates the
     *      distances of the child node to the entire scroll area of the
     *      container node.
     *
     * @returns {Object}
     *  An object with numeric properties representing the position and size of
     *  the child node relative to the entire scroll area or visible area of
     *  the container node, in pixels:
     *  - {Number} left
     *      The distance of the left border of the child node to the left
     *      border of the container node.
     *  - {Number} top
     *      The distance of the top border of the child node to the top border
     *      of the container node.
     *  - {Number} right
     *      The distance of the right border of the child node to the right
     *      border of the container node.
     *  - {Number} bottom
     *      The distance of the bottom border of the child node to the bottom
     *      border of the container node.
     *  - {Number} width
     *      The outer width of the child node (including its borders).
     *  - {Number} height
     *      The outer height of the child node (including its borders).
     */
    Utils.getChildNodePositionInNode = function (containerNode, childNode, options) {
        var pageRect = Utils.getNodePositionInPage(childNode);
        return Utils.getRectanglePositionInNode(containerNode, pageRect, options);
    };

    /**
     * Returns whether the specified child node is located inside the visible
     * area of its ancestor container node.
     *
     * @param {HTMLElement|jQuery} containerNode
     *  The DOM container element. If this object is a jQuery collection, uses
     *  the first node it contains.
     *
     * @param {HTMLElement|jQuery} childNode
     *  The DOM element whose visibility will be checked. Must be contained in
     *  the specified container element. If this object is a jQuery collection,
     *  uses the first node it contains.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.complete=false]
     *      If set to true, the child node must be located completely inside
     *      the visible area of the container node. Otherwise, it is sufficient
     *      that any part of the child node is visible in the container node.
     *
     * @returns {Boolean}
     *  Whether the specified child node is located inside the visible area of
     *  the container node.
     */
    Utils.isChildNodeVisibleInNode = function (containerNode, childNode, options) {
        var rectangle = Utils.getChildNodePositionInNode(containerNode, childNode, { visibleArea: true }),
            complete = Utils.getBooleanOption(options, 'complete', false);
        return complete ?
            ((rectangle.left >= 0) && (rectangle.right >= 0) && (rectangle.top >= 0) && (rectangle.bottom >= 0)) :
            ((rectangle.left + rectangle.width > 0) && (rectangle.right + rectangle.width > 0) && (rectangle.top + rectangle.height > 0) && (rectangle.bottom + rectangle.height > 0));
    };

    /**
     * Scrolls the passed page rectangle into the visible area of the specified
     * container node.
     *
     * @param {HTMLElement|jQuery} containerNode
     *  The DOM container element. If this object is a jQuery collection, uses
     *  the first node it contains. This method works independent from the
     *  current overflow mode of the container node.
     *
     * @param {Object} pageRect
     *  The document page rectangle that will be made visible by scrolling the
     *  container node. Must provide the properties 'left', 'top', 'width', and
     *  'height' in pixels. The properties 'left' and 'top' are interpreted
     *  relatively to the entire document page.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Number} [options.padding=0]
     *      Minimum distance between the borders of the visible area in the
     *      container node, and the rectangle scrolled into the visible area.
     *  @param {Boolean} [options.regardSoftkeyboard=false]
     *      special flag for upcoming softkeyboard, if set to true,
     *      the containerNodes clientheight will be interpreted smaller,
     *      that we better scroll to the visible area
     *  @param {Boolean} [options.forceToTop=false]
     *      Whether the specified pageRect shall be at the top border of the
     *      visible area of the scrolling node.
     */
    Utils.scrollToPageRectangle = function (containerNode, pageRect, options) {

        var // dimensions of the rectangle in the visible area of the scrollable node
            position = Utils.getRectanglePositionInNode(containerNode, pageRect, { visibleArea: true, regardSoftkeyboard: Utils.getBooleanOption(options, 'regardSoftkeyboard', false) }),
            // padding between scrolled element and border of visible area
            padding = Utils.getIntegerOption(options, 'padding', 0, 0);

        function updateScrollPosition(leadingChildOffset, trailingChildOffset, scrollAttributeName, opts) {

            var maxPadding = Utils.minMax((leadingChildOffset + trailingChildOffset) / 2, 0, padding),
                offset = Math.max(leadingChildOffset - Math.max(maxPadding - trailingChildOffset, 0), maxPadding);

            if (Utils.getBooleanOption(opts, 'forceToTop', false)) {
                Utils.getDomNode(containerNode)[scrollAttributeName] = leadingChildOffset + padding;
            } else {
                Utils.getDomNode(containerNode)[scrollAttributeName] -= (offset - leadingChildOffset);
            }
        }

        updateScrollPosition(position.left, position.right, 'scrollLeft');
        updateScrollPosition(position.top, position.bottom, 'scrollTop', { forceToTop: Utils.getBooleanOption(options, 'forceToTop', false) });
    };

    /**
     * Scrolls the passed child node into the visible area of the specified DOM
     * container node.
     *
     * @param {HTMLElement|jQuery} containerNode
     *  The DOM container element that contains the specified child node. If
     *  this object is a jQuery collection, uses the first node it contains.
     *
     * @param {HTMLElement|jQuery} childNode
     *  The DOM element that will be made visible by scrolling the specified
     *  container element. If this object is a jQuery collection, uses the
     *  first node it contains.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Number} [options.padding=0]
     *      Minimum distance between the borders of the visible area and the
     *      child node.
     *  @param {Boolean} [options.forceToTop=false]
     *      Whether the specified node shall be at the top border of the
     *      visible area of the scrolling node.
     */
    Utils.scrollToChildNode = function (containerNode, childNode, options) {
        var pageRect = Utils.getNodePositionInPage(childNode);
        Utils.scrollToPageRectangle(containerNode, pageRect, options);
    };

    /**
     * Sets the size of the passed DOM nodes. As a workaround for IE which
     * restricts the explicit size of single nodes to ~1.5m pixels (see comment
     * for the Utils.MAX_NODE_SIZE constant), descendant nodes will be inserted
     * whose sizes will add up to the specified total node size. The total size
     * of the node will still be restricted depending on the browser (see
     * comment for the Utils.MAX_CONTAINER_WIDTH and Utils.MAX_CONTAINER_HEIGHT
     * constants).
     *
     * @param {HTMLElement|jQuery} containerNode
     *  The DOM container element that will be resized. If this object is a
     *  jQuery collection, resizes the first node it contains.
     *
     * @param {Number} width
     *  The new total width of the node. The resulting width will not exceed
     *  the value of Utils.MAX_CONTAINER_WIDTH, depending on the current
     *  browser.
     *
     * @param {Number} height
     *  The new total height of the node. The resulting height will not exceed
     *  the value of Utils.MAX_CONTAINER_HEIGHT, depending on the current
     *  browser.
     *
     * @returns {Object}
     *  The effective size of the container node, in the properties 'width' and
     *  'height', in pixels (Internet Explorer does not report the correct
     *  total width and height of the container node if it exceeds the size
     *  limit of single nodes, as contained in Utils.MAX_NODE_SIZE).
     */
    Utils.setContainerNodeSize = function (containerNode, width, height) {

        var // the effective size of the container node (restrict to maximum allowed node size)
            resultSize = null;

        // convert to jQuery object, remove old helper nodes
        containerNode = $(containerNode).first();
        containerNode.find('>.ie-node-size-helper').remove();

        // restrict to maximum allowed node size
        width = Utils.minMax(width, 0, Utils.MAX_CONTAINER_WIDTH);
        height = Utils.minMax(height, 0, Utils.MAX_CONTAINER_HEIGHT);
        resultSize = { width: width, height: height };

        // if node is small enough, set its size directly
        if ((width <= Utils.MAX_NODE_SIZE) && (height <= Utils.MAX_NODE_SIZE)) {
            containerNode.css(resultSize);
            return resultSize;
        }

        // IE: insert embedded nodes to expand the container node beyond the limits of a single node

        // generate the mark-up for the embedded horizontal sizer nodes
        var markup = Utils.repeatString('<div style="width:' + Utils.MAX_NODE_SIZE + 'px;"></div>', Math.floor(width / Utils.MAX_NODE_SIZE));
        width %= Utils.MAX_NODE_SIZE;
        if (width > 0) { markup += '<div style="width:' + width + 'px;"></div>'; }

        // generate the mark-up for the vertical sizer nodes
        markup = '<div class="ie-node-size-helper" style="height:' + Math.min(height, Utils.MAX_NODE_SIZE) + 'px;">' + markup + '</div>';
        height = Math.max(0, height - Utils.MAX_NODE_SIZE);
        markup += Utils.repeatString('<div class="ie-node-size-helper" style="height:' + Utils.MAX_NODE_SIZE + 'px;"></div>', Math.floor(height / Utils.MAX_NODE_SIZE));
        height %= Utils.MAX_NODE_SIZE;
        if (height > 0) { markup += '<div class="ie-node-size-helper" style="height:' + height + 'px;"></div>'; }

        // insert entire HTML mark-up into the container node
        containerNode.css({ width: 'auto', height: 'auto' }).append(markup);

        // return the effective size of the container node
        return resultSize;
    };

    /**
     * Sets the absolute position of the passed DOM node in a container node.
     * As a workaround for IE which restricts the absolute position of a node
     * to ~1.5m pixels (see comment for the Utils.MAX_NODE_SIZE constant),
     * descendant embedded nodes will be inserted whose absolute positions will
     * add up to the specified total node position.
     *
     * @param {HTMLElement|jQuery} containerNode
     *  The DOM container element. If this object is a jQuery collection, uses
     *  the first node it contains.
     *
     * @param {HTMLElement|jQuery} childNode
     *  The DOM element that will be positioned absolutely in the container
     *  node. If this object is a jQuery collection, uses the first node it
     *  contains.
     *
     * @param {Number} left
     *  The new horizontal offset of the child node, in pixels.
     *
     * @param {Number} top
     *  The new vertical offset of the child node, in pixels.
     */
    Utils.setPositionInContainerNode = function (containerNode, childNode, left, top) {

        var // new helper nodes inserted into the container node
            helperNode = null,
            // current parent for the next helper node
            parentNode = null;

        // convert to jQuery objects
        containerNode = $(containerNode).first();
        childNode = $(childNode).first();

        // remove old helper nodes (detach first, child node may be embedded in old helper nodes)
        if (!childNode.parent().is(containerNode)) { childNode.detach(); }
        containerNode.find('>.ie-node-position-helper').remove();
        if (childNode.parent().length === 0) { containerNode.append(childNode); }

        // set position and size of the child node directly, if target position is valid
        if ((left <= Utils.MAX_NODE_SIZE) && (top <= Utils.MAX_NODE_SIZE)) {
            childNode.css({ position: 'absolute', left: left, top: top });
            return;
        }

        // create new helper nodes as long as the target position is beyond the node size limit
        parentNode = containerNode;
        while ((left > Utils.MAX_NODE_SIZE) || (top > Utils.MAX_NODE_SIZE)) {

            // create the new helper node
            helperNode = $('<div class="ie-node-position-helper">');

            // first helper node must be positioned explicitly
            if (parentNode === containerNode) {
                helperNode.css({ left: Math.min(left, Utils.MAX_NODE_SIZE), top: Math.min(top, Utils.MAX_NODE_SIZE) });
            }

            // reduce target position
            left = Math.max(0, left - Utils.MAX_NODE_SIZE);
            top = Math.max(0, top - Utils.MAX_NODE_SIZE);

            // set the size of the helper node (must not exceed node size limit)
            helperNode.css({ width: Math.min(left, Utils.MAX_NODE_SIZE), height: Math.min(top, Utils.MAX_NODE_SIZE) });

            // insert the new helper node into the DOM
            parentNode.append(helperNode);
            parentNode = helperNode;
        }

        // insert the child node into the last helper node, set its position to
        // the bottom-right corner of the last helper node
        childNode.css({ position: 'absolute', left: '100%', top: '100%' }).appendTo(parentNode);
    };

    // hidden storage in DOM --------------------------------------------------

    /**
     * Inserts new DOM nodes into a global hidden storage container. The nodes
     * will not be visible but will be part of the living DOM, thus it will be
     * possible to access and modify the geometry of the nodes.
     *
     * @param {HTMLElement|jQuery} nodes
     *  The DOM node(s) to be inserted into the private storage container. If
     *  this object is a jQuery collection, inserts all contained DOM nodes
     *  into the container.
     */
    Utils.insertHiddenNodes = function (nodes) {
        hiddenStorageNode.append(nodes);
    };

    /**
     * Returns the DOM nodes contained in the global hidden storage container.
     *
     * @param {String} selector
     *  The jQuery selector used to search for the contents in the global
     *  hidden storage node.
     *
     * @returns {jQuery}
     *  All matching DOM nodes, as jQuery collection.
     */
    Utils.findHiddenNodes = function (selector) {
        return hiddenStorageNode.find(selector);
    };

    // focus and selection ----------------------------------------------------

    /**
     * Returns whether one of the descendant elements in the passed element
     * contains the active (focused) element.
     *
     * @param {HTMLElement|jQuery} node
     *  A single DOM element, or a jQuery collection, whose descendants will be
     *  checked for the active (focused) element.
     *
     * @returns {Boolean}
     *  Whether one of the descendant elements in the passed element contains
     *  the active (focused) element.
     */
    Utils.containsFocus = function (node) {
        return $(node).find(window.document.activeElement).length > 0;
    };

    /**
     * Returns whether one of the descendant elements in the passed element, or
     * contains the active (focused) element; or the element itself is the
     * active element.
     *
     * @param {HTMLElement|jQuery} node
     *  A single DOM element, or a jQuery collection, that will be checked for
     *  the active (focused) element.
     *
     * @returns {Boolean}
     *  Whether one of the descendant elements in the passed element contains
     *  the active (focused) element, or the element itself is the active
     *  element.
     */
    Utils.hasOrContainsFocus = function (node) {
        return $(node).is(window.document.activeElement) || Utils.containsFocus(node);
    };

    /**
     * Move focus to from given node to its next or previous focusable node. (keyboard accessibility)
     *
     * @param {jQuery} startNode
     *  Starting node.
     *
     * @param {jQuery} focusableNode
     *  all focusable nodes
     *
     * @param {String = 'right'} direction
     *  Available direction strings: 'right','left', 'up', 'down'.
     *
     * @param {String|Object} [options]
     *  Optional parameters.
     *  @param {String|String[]} [options.noMoveAt]
     *      For example: 'input', 'textarea', ...
     * 		Single as a string, multiple as an array.
     * 		If 'startNode' is one of these TAGs, the caret-position (and 'direction')
     * 		decides to move the focus, or not (see 'stopMovingFocus').
     */
    Utils.moveNodeFocus = function (startNode, focusableNodes, direction, options) {

        if (!startNode) { return; }

        var nodeToFocus = startNode,
            length = focusableNodes.length,
            index = $(focusableNodes).index(startNode),
            direction = direction || 'right',
            noMoveAt = Utils.getOption(options, 'noMoveAt', null),
            returnVal = false;

       	function stopMovingFocus(node, direction){
       		var // Length of the node-value
       			len = $(node).val().length,
       			// Caret-position
        		pos = node[0].selectionStart;

        	if (pos === 0 && direction === 'left') {
        		// do nothing, let the caret jump to the prev control
        	} else if (pos === len && direction === 'right') {
        		// do nothing, let the caret jump to the next control
        	} else {
        		// stop moving the caret to the nex/prev control
        		// normal caret-traveling preferred (within the input-field)
        		return true;
        	}
       	}

       	// check (all) 'noMoveAt'-Tags against 'startNode'
        if (_.isArray(noMoveAt)) {
        	_.each(noMoveAt, function(noMoveEle){
        		if ($(startNode).is(noMoveEle)) {
		        	if(stopMovingFocus(startNode, direction)){ returnVal = true; }
		        }
        	});
        } else if (noMoveAt && $(startNode).is(noMoveAt)) {
        	if(stopMovingFocus(startNode, direction)){ returnVal = true; }
        }

        // if a 'noMoveAt'-Tag is focused and normal caret-traveling should
        // grab, get out of this method
        if (returnVal) { return; }

        switch (direction) {
            case 'right':
            case 'down':
                if (index < (length-1)) {
                    nodeToFocus = focusableNodes[(index+1)];
                }
                break;
            case 'left':
            case 'up':
                if (index > 0) {
                    nodeToFocus = focusableNodes[(index-1)];
                }
                break;
        }

        if (!nodeToFocus) { return; }

        nodeToFocus.focus();
        return nodeToFocus;
    };

    /**
     * Clears the current browser selection ranges.
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.rescueFocus=true]
     *      Internet Explorer only: If set to true, the focus will be moved to
     *      an internal helper DOM element before the browser selection will
     *      actually be cleared. This workaround is needed to prevent JS errors
     *      when accessing the browser selection of a hidden element. This time
     *      consuming action can be skipped via this option, if it is sure,
     *      that the browser selection points to visible DOM elements.
     */
    Utils.clearBrowserSelection = function (options) {
        try {
            // Bug 28515, bug 28711: IE fails to clear the selection (and to
            // modify it afterwards), if it currently points to a DOM node that
            // is not visible anymore. This happens e.g. after clicking on the
            // 'Show/hide side panel' button shown by all OX Documents
            // applications, which hide themselves together with the side pane
            // or overlay pane after activation. Workaround is to move focus to
            // an editable DOM node which will cause IE to update the browser
            // selection object. Using another focusable node (e.g. the body
            // element) is not sufficient. Use try/catch anyway to be notified
            // about other problems with the selection.
            if (_.browser.IE && Utils.getBooleanOption(options, 'rescueFocus', true)) {
                focusHelperNode.focus();
            }
            window.getSelection().removeAllRanges();
        } catch (ex) {
            Utils.exception(ex);
        }
    };

    // event handling ---------------------------------------------------------

    /**
     * Attaches an event handler to the given jQuery object and assures
     * that the handler is the first one to be executed if more than
     * one handler is attached to the object.
     *
     * Note: $._data() is not a supported public interface; the actual
     * data structures may change incompatibly from jQuery version to version.
     *
     * @param {jQuery} object
     *  The jQuery object the handler function is attached.
     * @param {String} event
     *  The event type and optional namespaces, such as 'click' or 'keydown.myPlugin'.
     * @param {Function} func
     *  The handler function to execute when the event is triggered.
     * @param {TriggerObject} [observer]
     *  If assigned via its listenTo-function the object.on() is called,
     *  and if observer is destroyed, listener gets removed
     * @returns {Boolean}
     *  Returns true if handler has been added as first one.
     *
     */
    Utils.bindAsFirstHandler = function (object, event, func, observer) {

        object = (object instanceof $) ? object : $(object);

        // normally attach the event first
        if (observer) {
            observer.listenTo(object, event, func);
        } else {
            object.on(event, func);
        }


        if (!_.isFunction($._data)) { return false; }

        // move the handler function
        object.each(function () {
            var // all handlers for the event
                handlers = $._data(this, 'events')[event.split('.')[0]],
                // the handler that has just been inserted at the end of the list
                handler = _.isArray(handlers) ? handlers.pop() : null;

            // move the handler to the beginning of the list
            if (handler) {
                handlers.splice(0, 0, handler);
            }
        });

        return true;
    };

    // debugging --------------------------------------------------------------

    /**
     * Starting the global timer.
     *
     * @param {String} message
     *  A message that is displayed in the console, when the timer starts.
     *
     * @param {Boolean} addTime
     *  Whether the time of several following global timers shall be accumulated.
     */
    Utils.startGlobalTimer = Config.DEBUG ? function (message, addTime) {
        if (message) { Utils.log(message); }
        accumulateTime = addTime || false;
        if ((accumulateTime) && (!_.isNumber(accumulatedTime))) {
            accumulatedTime = 0;
            accumulatedCounter = 0;
        }
        globalTimer = _.now();
    } : $.noop;

    /**
     * Logging info from the global timer.
     *
     * @param {String} message
     *  A message that is displayed in the console, if the timer runs.
     */
    Utils.logGlobalTimer = Config.DEBUG ? function (message) {
        if (globalTimer) {
            Utils.log(message + ' (' + (_.now() - globalTimer) + 'ms)');
        }
    } : $.noop;

    /**
     * Stopping the global timer.
     *
     * @param {String} message
     *  A message that is displayed in the console, when the timer is stopped.
     *
     * @returns {Number|Null}
     *  The measured time in milliseconds, or null, if no timer is running.
     */
    Utils.stopGlobalTimer = Config.DEBUG ? function (message) {
        var diff = null;
        if (globalTimer) {
            diff = _.now() - globalTimer;
            if (accumulateTime) {
                accumulatedTime += diff;
                accumulatedCounter++;
            }
            Utils.log(message + ' (' + diff + 'ms)' + (accumulateTime ? ' Accumulated time: ' + accumulatedTime + 'ms. Run: ' + accumulatedCounter : ''));
            globalTimer = null;
        }
        return diff;
    } : $.noop;

    /**
     * Checking if the global timer is already running.
     *
     * @returns {Boolean}
     *  Whether the global timer is running.
     */
    Utils.isGlobalTimerRunning = function () {
        return globalTimer !== null;
    };

    /**
     * Logging info from starting the global counter.
     *
     * @param {String} message
     *  A message that is displayed in the console, if the counter runs.
     */
    Utils.startGlobalCounter = Config.DEBUG ? function (message) {
        Utils.log(message);
        globalCounter = 0;
    } : $.noop;

    /**
     * Logging info from incrementing the global counter.
     *
     * @param {String} message
     *  A message that is displayed in the console, if the counter is incremented.
     */
    Utils.increaseGlobalCounter = Config.DEBUG ? function (message) {
        if (globalCounter !== null) {
            globalCounter++;
            if (message) {
                Utils.log(message + ' : ' + globalCounter);
            }
        }
    } : $.noop;

    /**
     * Logging info from stopping the global counter.
     *
     * @param {String} message
     *  A message that is displayed in the console, if the global counter is stopped.
     */
    Utils.stopGlobalCounter = Config.DEBUG ? function (message) {
        if (globalCounter !== null) {
            Utils.log(message + ' : ' + globalCounter);
            globalCounter = null;
        }
    } : $.noop;

    /**
     * A simple check to detect whether the local storage in the browser is available.
     * Problem: In Safari the local browsing mode does not support the usage of the
     * local storage.
     *
     * @returns {Boolean}
     *  Whether the local storage can be used.
     */
    Utils.isLocalStorageSupported = function () {

        var testKey = '_localTestKey_';

        try {
            localStorage.setItem(testKey, '1');
            localStorage.removeItem(testKey);
            return true;
        } catch (ex) {
            return false;
        }
    };

    // console output ---------------------------------------------------------

    /**
     * Writes a log message to the hidden Seleniun log node, if debug mode is
     * enabled in the global configuration.
     *
     * @param {Any} [...]
     *  The values to be written to the Selenium log.
     */
    Utils.logSelenium = Config.DEBUG ? (function () {

        var // console logger
            logger = new Logger({ enable: 'office:log-selenium', prefix: 'SELENIUM' }),
            // the global log node for Selenium tests
            logNode = $('<ul>', { id: 'io-ox-office-selenium' }).prependTo('body'),
            // global counter for log entries
            counter = 0;

        return function () {
            var args = _.toArray(arguments);
            counter += 1;
            args.unshift('id=' + counter);
            logger.info.apply(logger, args);
            logNode.append($('<li>').text(args.join(',')));
        };
    }()) : $.noop;



    /**
     * Changes the tabindex of the given node (OR their children)
     * e.g. to get nodes out of the tab-flow
     *
     * @param {jQuery} node
     *  The node, which should be changed
     *
     * @param {Object} [options]
     *  Optional parameters:
     *  @param {Boolean} [options.tabable=false]
     *      should change the tabindex to "1" (tabable), or to "-9" (not tabable)
     *  @param {Boolean} [options.childrenOfNode=false]
     *      change the given node (default), or change their children
     */
    Utils.changeTabState = function (node, options) {
    	var // node which should be changed
    		changeNode = null,
    		// set tabindex to
    		toTabindex = (Utils.getOption(options, 'tabable', false))?tabableTabindex:untabableTabindex,
    		// remove tabindex
    		fromTabindex = (Utils.getOption(options, 'tabable', false))?untabableTabindex:tabableTabindex,
    		// the given node, or their children
    		children = Utils.getOption(options, 'childrenOfNode', false);

    	changeNode = (children)?node.children():node;

    	changeNode.find('[tabindex="'+fromTabindex+'"]').attr('tabindex', toTabindex);
    	return true;
    };




    // global initialization ==================================================

    // insert the focus helper into the DOM
    Utils.insertHiddenNodes(focusHelperNode);

    // create global logger instances
    Utils.FocusLogger = new Logger({ enable: 'office:log-focus', prefix: 'FOCUS' });
    Utils.LayoutLogger = new Logger({ enable: 'office:log-layout', prefix: 'LAYOUT' });

    // focus tracker ----------------------------------------------------------

    // trigger normalized global 'change:focus' events
    (function () {

        var // the DOM node currently focused
            focusNode = null,
            // whether the last received event was a 'focusout' event
            pendingFocusOut = false,
            // last 'keydown' event received in this stack frame
            pendingKeyEvent = null;

        // update the current focus node, and trigger the global 'change:focus' event
        function updateFocus(newFocusNode) {
            var blurNode = focusNode;
            // ignore 'focusin' events on the document root (sent by Firefox when returning focus to browser tab)
            if ((document !== newFocusNode) && (focusNode !== newFocusNode)) {
                focusNode = newFocusNode;
                Utils.trigger('change:focus', focusNode, blurNode, pendingKeyEvent);
            }
        }

        // remember 'keydown' event object to be passed to the next 'change:focus' event
        function keyDownHandler(event) {
            pendingKeyEvent = event;
            _.defer(function () { pendingKeyEvent = null; });
        }

        // handle a new focused DOM element
        function focusInHandler(event) {
            Utils.FocusLogger.log('type="focus"', 'target=', event.target, 'active=', document.activeElement);
            pendingFocusOut = false;
            updateFocus(event.target);
        }

        // handle a 'focusout' event, detect whether entire page (browser tab) has lost focus
        function focusOutHandler(event) {
            Utils.FocusLogger.log('type="blur"', 'target=', event.target, 'active=', document.activeElement);
            // ignore 'focusout' event whose target does not match the current focus node
            if (event.target === focusNode) {
                pendingFocusOut = true;
                _.defer(function () {
                    if (pendingFocusOut) {
                        pendingFocusOut = false;
                        updateFocus(null);
                    }
                });
            }
        }

        // listen to 'keydown' events in the capturing phase, otherwise custom focus handlers
        // will be executed before the 'keydown' event bubbles up to the document node
        document.addEventListener('keydown', keyDownHandler, true);

        // listen to the global bubbling 'focusin' and 'focusout' events of jQuery
        $(document).on({ focusin: focusInHandler, focusout: focusOutHandler });

        // log focus events
        Utils.FocusLogger.logEvent(Utils, 'change:focus', 'focus', 'blur', 'keyboard');

        // initialize current focus node by calling the 'focusin' event handler directly
        focusInHandler({ target: document.activeElement });
    }());

    // exports ================================================================

    return Utils;

});
